unit estime;

interface

uses
  SysUtils, Types, Classes, QGraphics, QControls, QForms, QDialogs,
  QStdCtrls;

type
  TFEstime = class(TForm)
    Button1: TButton;
    Button2: TButton;
    Label1: TLabel;
    Edit1: TEdit;
    Label2: TLabel;
    Edit2: TEdit;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  FEstime: TFEstime;

implementation

uses
  fonctions, rapport, main, optim, matrices, vision;

{$R *.xfm}

procedure TFEstime.Button1Click(Sender: TObject);
var
  i,j, k, MaxIter: Integer;
  str: string;
  tolransac: Real;
begin
  { tester les pts d'intrt }
  if (Length(PtsIntApp) = 0) then begin
    MessageDlg('Des points d''intrt apparis sont ncessaires pour cette tape.', mtError, [mbOK], 0);
    Exit;
  end;

  { tester les valeurs entrees }
  try
    MaxIter:= StrToInt(Edit1.Text);
  except
    MessageDlg('Le nombre d''itrations n''est pas une valeur correcte.', mtError, [mbOK], 0);
    Exit;
  end;

  try
    tolransac:= StrToFloat(Edit2.Text);
  except
    MessageDlg('La tolrance ransac n''est pas une valeur correcte.', mtError, [mbOK], 0);
    Exit;
  end;

  FRapport.Show;
  FRapport.Write('====================================================');
  FRapport.Write('=========== Estimation des homographies ==============');
  FRapport.Write('====================================================');
  FRapport.Write('  Nombre maximal d''itrations pour l''estimation linaire : '+Edit1.Text);
  FRapport.Write('  Tolrance ransac : '+Edit2.Text);
  FRapport.Write('');
  EstimeH(MaxIter, tolransac);
  for k:= 0 to ImageG.Polygones.Count-1 do begin
    if (Polygone(ImageG.Polygones[k]).Locked) then
      Continue;
    FRapport.Write('  --> Polygone '+IntToStr(k+1)+'/'+IntToStr(ImageG.Polygones.Count));
    FRapport.Write('        Estimation linaire: H =');
    str:= '        [ ';
    for i:= 0 to 2 do begin
      for j:= 0 to 2 do begin
        str:= str + Format('%10.8f', [Hom[k][i][j]])+' ';
      end;
      if (i = 2) then
        str:= str + ' ]'
      else
        str:= str + #10#13+'          ';
    end;
    FRapport.Write(str);

    FRapport.Write(' ');
    FRapport.Write('        Estimation non linaire: H =');
    str:= '        [ ';
    for i:= 0 to 2 do begin
      for j:= 0 to 2 do begin
        str:= str + Format('%10.8f', [HomN[k][i][j]])+' ';
      end;
      if (i = 2) then
        str:= str + ' ]'
      else
        str:= str + #10#13+'          ';
    end;
    FRapport.Write(str);
    FRapport.Write(' ');
    FRapport.Write('        Valeur de la fonction objective avant optimisation: '+FloatToStr(Obj[k][0]));
    FRapport.Write('        Valeur de la fonction objective aprs optimisation: '+FloatToStr(Obj[k][1]));
    if (Obj[k][2] <> OPT_OK) then begin
      if (Obj[k][2] = OPT_SING) then
        FRapport.Write('        ******** Matrice hessienne singulire ! *********')
      else if (Obj[k][2] = OPT_NON_CONV) then
        FRapport.Write('        ******** Non convergence ! ********')
      else if (Obj[k][2] = OPT_BIG_LAMBDA) then
        FRapport.Write('        ******** Paramtre de Marquardt trop grand ! ********')
      else if (Obj[k][2] = MAT_NON_CONV) then
        FRapport.Write('        ******** Non convergence de la dcomposition SVD ! ********');
    end
    else
      FRapport.Write('        ******** Convergence ! ***********');
    FRapport.Write(' ');



    FRapport.Write('        Estimation robuste: H =');
    str:= '        [ ';
    for i:= 0 to 2 do begin
      for j:= 0 to 2 do begin
        str:= str + Format('%10.8f', [HomR[k][i][j]])+' ';
      end;
      if (i = 2) then
        str:= str + ' ]'
      else
        str:= str + #10#13+'          ';
    end;
    FRapport.Write(str);
    FRapport.Write(' ');
    FRapport.Write('        nombre de points : '+IntToStr(Length(PtsIntApp[k])));
    FRapport.Write('        nombre d''inliers : '+FloatToStr(Obj[k][3]));
    if (Obj[k][4] <> RAN_OK) then begin
      if (Obj[k][4] = RAN_DEG) then
        FRapport.Write('        ******** Impossible de trouver une configuration non dgnre ! *********')
      else if (Obj[k][4] = RAN_NO) then
        FRapport.Write('        ******** Ransac n''a pas trouv de solution ! ********')
      else if (Obj[k][4] = RAN_MAX) then
        FRapport.Write('        ******** Ransac a atteint le nombre maximal d''itrations ! ********');
    end
    else
      FRapport.Write('        ******** Ransac a trouv une solution ! ***********');
    FRapport.Write(' ');

    FRapport.Write(' ');
    FRapport.Write('        Estimation non linaire robuste : H =');
    str:= '        [ ';
    for i:= 0 to 2 do begin
      for j:= 0 to 2 do begin
        str:= str + Format('%10.8f', [HomRN[k][i][j]])+' ';
      end;
      if (i = 2) then
        str:= str + ' ]'
      else
        str:= str + #10#13+'          ';
    end;
    FRapport.Write(str);
    FRapport.Write(' ');
    FRapport.Write('        Valeur de la fonction objective avant optimisation: '+FloatToStr(Obj[k][5]));
    FRapport.Write('        Valeur de la fonction objective aprs optimisation: '+FloatToStr(Obj[k][6]));
    if (Obj[k][7] <> OPT_OK) then begin
      if (Obj[k][7] = OPT_SING) then
        FRapport.Write('        ******** Matrice hessienne singulire ! *********')
      else if (Obj[k][7] = OPT_NON_CONV) then
        FRapport.Write('        ******** Non convergence ! ********')
      else if (Obj[k][7] = OPT_BIG_LAMBDA) then
        FRapport.Write('        ******** Paramtre de Marquardt trop grand ! ********')
      else if (Obj[k][7] = MAT_NON_CONV) then
        FRapport.Write('        ******** Non convergence de la dcomposition SVD ! ********');
    end
    else
      FRapport.Write('        ******** Convergence ! ***********');
    FRapport.Write(' ');

  end;
  FRapport.Write(' ');
  FRapport.Write(' ');
  //MessageDlg('L''estimation des homographies est termine.', mtInformation, [mbOK], 0);
  Close;
end;

procedure TFEstime.Button2Click(Sender: TObject);
begin
  Close;
end;

end.
