unit fonctions;

interface

uses
  SysUtils, QGraphics, QDialogs, Classes, types, matrices, fmath;

type
  TMatInt = array of array of Integer;
  pMatInt = ^TMatInt;

type
  TRPoint = record
    ind: Integer;
    val: Real;
    U, V: Real;
    i, j: Real;
  end;

type
  TRoint = record
    x, y: Real;
  end;

type
  ATRPoint = array of TRPoint;

type
  TRRect = packed record
    case Integer of
      0: (Left, Top, Right, Bottom: Real);
      1: (TopLeft, BottomRight: TRoint);
  end;


type
  Sommet = class
    private
    public
      index: Integer;
      U, V: Real;
      BordU, BordV, Bloque: Boolean;
      constructor Create(U, V: Real); overload;
      constructor Create; overload;
      function Clone: Sommet;
  end;

  Polygone = class
    private
    public
      Sommets: TList;
      Energie: Real;
      Locked: Boolean;
      constructor Create;
      destructor Destroy; override;
      procedure FreeSommets;
      procedure Clone(p: Polygone);
      procedure CopyUVFrom(p: Polygone);
      procedure SetLocked(locked: Boolean);
  end;

  type
  pRGBQuad = ^TRGBQuad;
  TRGBQuad =
  PACKED RECORD
    rgbBlue : BYTE;
    rgbGreen: BYTE;
    rgbRed : BYTE;
    rgbReserved: BYTE
  END;
  TYPE TRGBQuadArray = ARRAY[WORD] OF TRGBQuad;
  pRGBQuadArray = ^TRGBQuadArray;

  PIntArray = ^TIntArray;
  TIntArray = array[0..16383] of Integer;

var
  E_Segments: array of array of Real;
  E_Segments_Index: Integer;
  NN: Integer;

const

  SOBELX: array[0..8] of Integer = (-1,0,1,-2,0,2,-1,0,1);
  SOBELY: array[0..8] of Integer = (-1,-2,-1,0,0,0,1,2,1);
  TAILLE_FEN = 1;
  INTEMAX = 100000;
  TOL = 1.0E-8;
  DE3 = 256*256*256;
  DE2 = 256*256;
  DE1 = 256;

{ calcule le produit vectoriel de deux vecteurs 2D }
procedure ProdVect(v, w: array of Real; var z: array of Real);

{ 'normalise' un vecteur 2D homogene }
procedure NormaliseHomogene2D(var v: array of Real);

{ calcule le gradient de l'image 'image' avec le filtre de convolution
  'mask' puis tend le gradient d'une ligne et d'une colonne pour les
  interpolations bicubiques }
function Gradient(image: TBitmap; const mask: array of integer): pMatInt;

{ chantillone un segment avec un pas maximal de 'pas' }
function EchantSegment(Sommet1, Sommet2: Sommet; pas: Real): TList;

{ concatne la liste 'liste2'  la liste 'liste1' }
procedure ConcatListe(Liste1, Liste2: TList);

{ tri des sommets par U ou V croissants ou dcroissants }
function TriUC(Item1, Item2: Pointer): Integer;
function TriVC(Item1, Item2: Pointer): Integer;
function TriUD(Item1, Item2: Pointer): Integer;
function TriVD(Item1, Item2: Pointer): Integer;

{ calcule la norme d'un vecteur }
function NormVect(Sommet1, Sommet2: Sommet): Real;

{ calcule l'nergie d'un segment }
function EnergieSegment(points: TList; pgradx, pgrady: pMatInt): Real;

{ calcule l'energie d'un polygone }
function EnergiePoly(polys: Polygone; pas: Real; gradimx, gradimy: pMatInt): Real;

{ calcule l'interpolation bicubique de (u,v) dans pmat }
function InterpBicubique(u, v: Real; pmat: pMatInt): Real;

{ parcours un niveau de profondeur dans l'espace des polygones }
procedure ParcoursEspace(niveau, niveau_max: Integer; imgradx, imgrady: pMatInt; polys, polyi: Polygone; taille: Real; pas: Real; Emax: Real; polyf: Polygone);

{ cherche le meilleur polygone dans tout l'espace }
function polySnake2(imgradx, imgrady: pMatInt; polyf: Polygone; pas: Real; nite: Integer; profondeur: array of Integer; nsom, maxiter: Integer): Polygone;

{ fonctions de stockage, recherche et gestions des energies de segments }
procedure InitE_Segments(N: Integer);
procedure AugmenteE_Segments(N: Integer);
procedure StockeE_Segments(u1, v1, u2, v2, E: Real);
function ChercheE_Segments(u1, v1, u2, v2: Real): Real;
procedure VideE_Segments;

{ dtecte des points d'intrt dans l'image de gauche }
procedure DetectPts(polygones: TList; TheImage: TBitmap; N, N_poly: Integer; Npoints: Integer);

{ retourne le min et l'indice du min d'un tableau }
function MinTab(tab: array of TRPoint): TRPoint;

{ retourne le max et l'indice du max d'un tableau }
function MaxTab(tab: array of TRPoint): TRPoint;

{ retourne le max local si il existe dans une zone n2 x n2 }
function MaxLocalTab(tab: array of TRPoint; U, V: Real; n2: Integer): TRPoint;

{ apparie des pixels entre les deux images }
procedure Apparie(n,p1,p2,q1,q2: Integer; ccncMin: Real; prec: Real; bidir: Integer);

{ apparie un pixel au sous pixel }
function SubCorresp(Z1,Z2: TRect; prec: Real; bidir, p1, p2, q1, q2: Integer): TRPoint;

{ apparie un pixel au pixel prs }
function Corresp(Z1,Z2: TRect; ImDep, ImArr: TBitmap; polyArr: TList): TRPoint;

{ calcule un ccnc entre 2 tableaux }
function CCNC(f1, f2: ATRPoint): Real;

{ estime les Homographies }
procedure EstimeH(MaxIter: Integer; tolransac: Real);

{ pour dbugger }
procedure WriteMatrix(Title : String; A : TMatrix; N, M : Integer);
procedure WriteVector(Title : String; A : TVector; N: Integer);

{ calcule les disparits et localise les occultations }
procedure CalcDisp(Facteur, mult, prec, homo: Integer);

{ la fonction objective  optimiser pour l'estimation des homographies }
function FObj(X : TVector) : Float;

{ la fonction objective  optimiser pour l'optimisation de la segmentation }
function FObjPoly(X : TVector) : Float;

{ fonction pour savoir si un point est dans un polygone }
function InPolygon(poly: Polygone; u,v: Real): Boolean;

{ fonction pour savoir dans quel polygone est un point }
function FindPolygon(polys: TList; u,v: Real): Integer;

{ TPoint et TRect en real }
function RRect(Left, Top, Right, Bottom: Real): TRRect;
function Roint(X, Y: Real): TRoint;
function PtInRRect(const Rect: TRRect; const P: TRoint): Boolean;

function CheckPolys(polys: TList): Boolean;

implementation

uses main, math, rapport, polyfunc, optim, simopt, vision;

var
  polyEval: Polygone;
  pasEval: Real;
  gradimxEval, gradimyEval: pMatInt;

procedure ProdVect(v, w: array of Real; var z: array of Real);
begin
  z[0]:= v[1]*w[2]-v[2]*w[1];
  z[1]:= v[2]*w[0]-v[0]*w[2];
  z[2]:= v[0]*w[1]-v[1]*w[0];
end;

procedure NormaliseHomogene2D(var v: array of Real);
begin
  v[0]:= v[0] / v[2];
  v[1]:= v[1] / v[2];
  v[2]:= 1;
end;

procedure Polygone.SetLocked(locked: Boolean);
var
  i: Integer;
begin
  Self.Locked:= locked;
  for i:= 0 to Sommets.Count-1 do
    Sommet(Sommets[i]).Bloque:= locked;
end;

function CheckPolys(polys: TList): Boolean;
var
  i: Integer;
  poly: Polygone;
begin
  Result:= true;
  for i:= 0 to polys.Count-1 do begin
    poly:= Polygone(polys[i]);
    if (poly.Sommets[0] <> poly.Sommets[poly.Sommets.Count-1]) then begin
      Result:= false;
      Exit;
    end;
  end;
end;

function RRect(Left, Top, Right, Bottom: Real): TRRect;
begin
  Result.Left := Left;
  Result.Top := Top;
  Result.Bottom := Bottom;
  Result.Right := Right;
end;

function Roint(X, Y: Real): TRoint;
begin
  Result.X := X;
  Result.Y := Y;
end;

function PtInRRect(const Rect: TRRect; const P: TRoint): Boolean;
begin
  Result := (P.X >= Rect.Left) and (P.X <= Rect.Right) and (P.Y >= Rect.Top)
    and (P.Y <= Rect.Bottom);
end;

function FindPolygon(polys: TList; u,v: Real): Integer;
var
  i: Integer;
begin
  Result:= -1;
  for i:= 0 to polys.Count-1 do begin
    if (InPolygon(polys[i], u, v)) then begin
      Result:= i;
      Exit;
    end;
  end;
end;

function InPolygon(poly: Polygone; u,v: Real): Boolean;
var
  points: array of TRoint;
  i: Integer;
begin
  SetLength(points, poly.Sommets.Count-1);
  for i:= 0 to Length(points)-1 do
    points[i]:= Roint(Sommet(poly.Sommets[i]).V, Sommet(poly.Sommets[i]).U);
  Result:= G_PtInPolygon(points, Roint(v, u));
  points:= nil;
end;

function FObjPoly(X : TVector) : Float;
var
  l: Integer;
  som: Sommet;
begin
  { construire le polygone }
  for l:= 0 to polyEval.Sommets.Count-2 do begin
    som:= polyEval.Sommets[l];
    som.U:= Round(X[2*l]*10)/10;
    som.V:= Round(X[2*l+1]*10)/10;
  end;
  som:= polyEval.Sommets[polyEval.Sommets.Count-1];
  som.U:= Round(X[0]*10)/10;
  som.V:= Round(X[1]*10)/10;
  { calcule de son nergie }
  Result:= 1/EnergiePoly(polyEval, pasEval, gradimxEval, gradimyEval);
end;

function FObj(X : TVector) : Float;
var
  x147, x258, x369, yu, yv: Real;
  i, k: Integer;
begin
  Result:= 0;
  k:= KK;
  for i:= 0 to Length(PtsIntApp[k])-1 do begin
    x147:= X[0]*PtsIntApp[k][i].i + X[3]*PtsIntApp[k][i].j + X[6];
    x258:= X[1]*PtsIntApp[k][i].i + X[4]*PtsIntApp[k][i].j + X[7];
    x369:= X[2]*PtsIntApp[k][i].i + X[5]*PtsIntApp[k][i].j + X[8];
    yu:= PtsIntApp[k][i].U - x147/x369;
    yv:= PtsIntApp[k][i].V - x258/x369;
    Result:= Result + sqrt(yu*yu+yv*yv);
  end;
end;

procedure CalcDisp(Facteur, mult, prec, homo: Integer);
var
  i,j, n, ip, jp, l, n_sspix: Integer;
  k, kp, kp2, k1, k2: Integer;
  u,v,w,d,up,vp,du,dv: Real;
  di: Integer;
  rowsc, rowscv, rowscvg: pRGBQuadArray;
  rowsocc: pRGBQuadArray;
  rowsocca: array of pRGBQuadArray;
  rowsca: array of pRGBQuadArray;
  Carte, Carte1, Carte1G, CarteOcc, Carte1Occ: TBitmap;
  PolyPoints: array of TRoint;
  ptTest: TRoint;
  poly: Polygone;
  disp: array of TRPoint;
  res: TRPoint;
  epsi: Real;
  valeurs: TStrings;
  s: string;
begin
  epsi:= 1e-5;
  Carte:= TBitmap.Create;
  Carte.Width:= ImageG.TheImage.Width*Facteur;
  Carte.Height:= ImageG.TheImage.Height*Facteur;
  Carte.PixelFormat:= pf32bit;

  CarteOcc:= TBitmap.Create;
  CarteOcc.Width:= ImageG.TheImage.Width*Facteur;
  CarteOcc.Height:= ImageG.TheImage.Height*Facteur;
  CarteOcc.PixelFormat:= pf32bit;
  n:= 0;

  for i:= 0 to Carte.Height-1 do begin
    rowsc:= Carte.ScanLine[i];
    rowsocc:= CarteOcc.ScanLine[i];
    for j:= 0 to Carte.Width-1 do begin
      rowsocc[j].rgbBlue:= 0;
      rowsocc[j].rgbGreen:= 0;
      rowsocc[j].rgbRed:= 0;
      rowsocc[j].rgbReserved:= 0;

      u:= i / Facteur;
      v:= j / Facteur;
      k:= FindPolygon(ImageG.Polygones, u, v);
      if (k <> -1) then
        if (Polygone(ImageG.Polygones[k]).Locked) then
          k:= -1;
      if ((k < ImageG.Polygones.Count) and (k >= 0)) then begin
        up:= 0;
        vp:= 0;
        if (homo = 0) then begin
          up:= Hom[k][0][0]*u + Hom[k][0][1]*v + Hom[k][0][2];
          vp:= Hom[k][1][0]*u + Hom[k][1][1]*v + Hom[k][1][2];
          w:= Hom[k][2][0]*u + Hom[k][2][1]*v + Hom[k][2][2];
        end
        else if (homo = 1) then begin
          up:= HomN[k][0][0]*u + HomN[k][0][1]*v + HomN[k][0][2];
          vp:= HomN[k][1][0]*u + HomN[k][1][1]*v + HomN[k][1][2];
          w:= HomN[k][2][0]*u + HomN[k][2][1]*v + HomN[k][2][2];
        end
        else if (homo = 2) then begin
          up:= HomR[k][0][0]*u + HomR[k][0][1]*v + HomR[k][0][2];
          vp:= HomR[k][1][0]*u + HomR[k][1][1]*v + HomR[k][1][2];
          w:= HomR[k][2][0]*u + HomR[k][2][1]*v + HomR[k][2][2];
        end
        else if (homo = 3) then begin
          up:= HomRN[k][0][0]*u + HomRN[k][0][1]*v + HomRN[k][0][2];
          vp:= HomRN[k][1][0]*u + HomRN[k][1][1]*v + HomRN[k][1][2];
          w:= HomRN[k][2][0]*u + HomRN[k][2][1]*v + HomRN[k][2][2];
        end;

        up:= up / w;
        vp:= vp / w;
        dv:= vp - v;
        du:= up - u;
        d:= sqrt(du*du + dv*dv);
        ip:= Round(up*Facteur);
        jp:= Round(vp*Facteur);
        if (ip <> i) then
          Inc(n);
        { le correspondant est hors de l'image droite }
        {
        if ((ip < -epsi) or (ip >= (ImageD.TheImage.Height*Facteur)) or (jp < -epsi) or (jp >= (ImageD.TheImage.Width*Facteur))) then begin
          rowsc[j].rgbBlue:= 0;
          rowsc[j].rgbGreen:= 0;
          rowsc[j].rgbRed:= 0;
          rowsc[j].rgbReserved:= 0;
          rowsocc[j].rgbBlue:= 255;
          rowsocc[j].rgbGreen:= 255;
          rowsocc[j].rgbRed:= 255;
          rowsocc[j].rgbReserved:= 0;
        end
        else begin
        }
          kp:= FindPolygon(ImageD.Polygones, up, vp);
          if (kp = -1) then begin
            rowsc[j].rgbBlue:= 0;
            rowsc[j].rgbGreen:= 0;
            rowsc[j].rgbRed:= 0;
            rowsc[j].rgbReserved:= 0;
            rowsocc[j].rgbBlue:= 255;
            rowsocc[j].rgbGreen:= 255;
            rowsocc[j].rgbRed:= 255;
            rowsocc[j].rgbReserved:= 0;
          end
          else if (k <> kp) then begin
            rowsc[j].rgbBlue:= 0;
            rowsc[j].rgbGreen:= 0;
            rowsc[j].rgbRed:= 0;
            rowsc[j].rgbReserved:= 0;
            rowsocc[j].rgbBlue:= 255;
            rowsocc[j].rgbGreen:= 255;
            rowsocc[j].rgbRed:= 255;
            rowsocc[j].rgbReserved:= 0;
          end
          else begin
            di:= Round(d*prec);
            rowsc[j].rgbBlue:= Floor(di/DE2);
            di:= di mod DE2;
            rowsc[j].rgbGreen:= Floor(di/DE1);
            di:= di mod DE1;
            rowsc[j].rgbRed:= di;
            rowsc[j].rgbReserved:= 0;
          end;
        //end;
      end
      else begin
        rowsc[j].rgbBlue:= 0;
        rowsc[j].rgbGreen:= 0;
        rowsc[j].rgbRed:= 0;
        rowsc[j].rgbReserved:= 0;
      end;
    end;
  end;
  Carte.SaveToFile('carte.bmp');

  { si la rsolution est plus petite que 1 alors on calcule une carte
    au pixel prs en moyennant la carte calcule }
  //if (Facteur <> 1) then begin
  valeurs:= TStringList.Create;

    Carte1:= TBitmap.Create;
    Carte1.Width:= ImageG.TheImage.Width;
    Carte1.Height:= ImageG.TheImage.Height;
    Carte1.PixelFormat:= pf32bit;
    Carte1G:= TBitmap.Create;
    Carte1G.Width:= ImageG.TheImage.Width;
    Carte1G.Height:= ImageG.TheImage.Height;
    Carte1G.PixelFormat:= pf32bit;
    Carte1Occ:= TBitmap.Create;
    Carte1Occ.Width:= ImageG.TheImage.Width;
    Carte1Occ.Height:= ImageG.TheImage.Height;
    Carte1Occ.PixelFormat:= pf32bit;
    SetLength(rowsca, Facteur);
    SetLength(rowsocca, Facteur);
    for i:= 0 to ImageG.TheImage.Height-1 do begin
      rowscv:= Carte1.ScanLine[i];
      rowscvg:= Carte1G.ScanLine[i];
      rowsocc:= Carte1Occ.ScanLine[i];
      s:= '';

      { on rcupre toutes les sous-lignes incluses dans cette ligne d'paisseur
        1 pixel }
      for k:= 0 to Facteur-1 do begin
        rowsca[k]:= Carte.ScanLine[i*Facteur+k];
        rowsocca[k]:= CarteOcc.ScanLine[i*Facteur+k];
      end;

      for j:= 0 to ImageG.TheImage.Width-1 do begin
        { on moyenne les disparits et on compte les sous-pixels occults dans
          la zone de taille Facteur*Facteur sous pixels }
        n:= 0;
        SetLength(disp, ImageG.Polygones.Count+1);
        for k2:= 0 to Length(disp)-1 do begin
          disp[k2].val:= 0;
          disp[k2].u:= 0;
          disp[k2].ind:= k2;
        end;
        for k:= 0 to Facteur-1 do begin
          for l:= 0 to Facteur-1 do begin
            if ((rowsca[k][j*Facteur+l].rgbBlue = 0) and (rowsca[k][j*Facteur+l].rgbGreen = 0)
             and (rowsca[k][j*Facteur+l].rgbRed = 0) and (rowsca[k][j*Facteur+l].rgbReserved = 0)) then begin
              disp[ImageG.Polygones.Count].val:= disp[ImageG.Polygones.Count].val + 1;
            end
            else begin
              kp:= FindPolygon(ImageG.Polygones, i+k/Facteur, j+l/Facteur);
              if (kp = -1) then
                disp[ImageG.Polygones.Count].val:=  disp[ImageG.Polygones.Count].val + 1
              else begin
                disp[kp].val:=  disp[kp].val + 1;
                disp[kp].u:= disp[kp].u + rowsca[k][j*Facteur+l].rgbBlue*DE2 + rowsca[k][j*Facteur+l].rgbGreen*DE1 + rowsca[k][j*Facteur+l].rgbRed;
              end;
            end;
            { carte des occultations }
            if ((rowsocca[k][j*Facteur+l].rgbBlue = 255) and (rowsocca[k][j*Facteur+l].rgbGreen = 255)
             and (rowsocca[k][j*Facteur+l].rgbRed = 255) and (rowsocca[k][j*Facteur+l].rgbReserved = 0)) then
              Inc(n);
          end;
        end;
        { trouver la population majoritaire }
        res:= MaxTab(disp);

        if (res.ind = (Length(disp)-1)) then begin
          rowscv[j].rgbBlue:= 0;
          rowscv[j].rgbGreen:= 0;
          rowscv[j].rgbRed:= 0;
          rowscv[j].rgbReserved:= 0;
          rowscvg[j].rgbBlue:= 0;
          rowscvg[j].rgbGreen:= 0;
          rowscvg[j].rgbRed:= 0;
          rowscvg[j].rgbReserved:= 0;
          rowsocc[j].rgbBlue:= 255;
          rowsocc[j].rgbGreen:= 255;
          rowsocc[j].rgbRed:= 255;
          rowsocc[j].rgbReserved:= 0;
          s:= s + '0 '; 
        end
        else begin
          res.U:= disp[res.ind].U;
          s:= s + FloatToStr(res.U/res.val/prec)+' ';
          di:= Round(res.U*mult/res.val/prec);
          rowscvg[j].rgbBlue:= di;
          rowscvg[j].rgbGreen:= di;
          rowscvg[j].rgbRed:= di;
          rowscvg[j].rgbReserved:= 0;

          di:= Round(res.U/res.val);
          rowscv[j].rgbBlue:= Floor(di/DE2);
          di:= di mod DE2;
          rowscv[j].rgbGreen:= Floor(di/DE1);
          di:= di mod DE1;
          rowscv[j].rgbRed:= di;
          rowscv[j].rgbReserved:= 0;

          rowsocc[j].rgbBlue:= 0;
          rowsocc[j].rgbGreen:= 0;
          rowsocc[j].rgbRed:= 0;
          rowsocc[j].rgbReserved:= 0;
        end;
      end;
      valeurs.Add(s);
    end;
    valeurs.SaveToFile('disparites.txt');
    valeurs.Free;
    Carte1.SaveToFile('carte1.bmp');
    Carte1.Free;
    Carte1G.SaveToFile('carte1gris.bmp');
    Carte1G.Free;
    Carte1Occ.SaveToFile('carte1occ.bmp');
    Carte1Occ.Free;
  //end;

  Carte.Free;
  CarteOcc.Free;
end;

procedure WriteVector(Title : String; A : TVector; N: Integer);
{ ----------------------------------------------------------------------
  Writes a vector A (N)
  ---------------------------------------------------------------------- }
  var
    I: Integer;
  begin
    FRapport.Write(Title+' :');
    FRapport.Write(' ');
    for I := 0 to N-1 do
      begin
          FRapport.Write(FloatToStr(A[I])+' ');
      end;
    FRapport.Write(' ');
  end;

  procedure WriteMatrix(Title : String; A : TMatrix; N, M : Integer);
{ ----------------------------------------------------------------------
  Writes a matrix A (N x M)
  ---------------------------------------------------------------------- }
  var
    I, J : Integer;
  begin
    FRapport.Write(Title+' :');
    FRapport.Write(' ');
    for I := 0 to N-1 do
      begin
        for J := 0 to M-1 do
          FRapport.Write(FloatToStr(A[I,J])+' ');
        FRapport.Write(' ');
      end;
    FRapport.Write(' ');
  end;

procedure EstimeH(MaxIter: Integer; tolransac: Real);
var
  i,j,k, l, N: Integer;
  res: Integer;
  pts1, pts2, H, Hn, ptsin1, ptsin2: TMatrix;
  fi, ff: Real;
  ninliers: Integer;
begin
  SetLength(Hom, Length(PtsIntApp));
  SetLength(HomN, Length(PtsIntApp));
  SetLength(HomR, Length(PtsIntApp));
  SetLength(HomRN, Length(PtsIntApp));
  SetLength(Obj, Length(PtsIntApp));
  SetLength(PtsIntAppRob, ImageG.Polygones.Count);

  for k:= 0 to Length(PtsIntApp)-1 do begin
    if (Polygone(ImageG.Polygones[k]).Locked) then
      Continue;
    N:= Length(PtsIntApp[k]);
    DimMatrix(pts1, 2, N-1);
    DimMatrix(pts2, 2, N-1);
    for l:= 0 to N-1 do begin
      pts1[0][l]:= PtsIntApp[k][l].i;
      pts1[1][l]:= PtsIntApp[k][l].j;
      pts1[2][l]:= 1;
      pts2[0][l]:= PtsIntApp[k][l].u;
      pts2[1][l]:= PtsIntApp[k][l].v;
      pts2[2][l]:= 1;
    end;

    DimMatrix(H, 2, 2);
    res:= homography2d(pts1, pts2, H, TOL);
    if (res <> MAT_OK) then begin
      for i:= 0 to 2 do
        for j:= 0 to 2 do
          HomN[k][i][j]:= 0;
      for i:= 0 to 2 do
        for j:= 0 to 2 do
          Hom[k][i][j]:= 0;
      Obj[k][0]:= 0;
      Obj[k][2]:= MAT_NON_CONV;
      Obj[k][1]:= 0;
      Continue;
    end;

    for i:= 0 to 2 do
      for j:= 0 to 2 do
        Hom[k][i][j]:= H[i][j];

    { estimation non linaire }
    DimMatrix(Hn, 2, 2);

    res:= homography2dnlin(pts1, pts2, H, Hn,maxiter,fi, ff);
    Obj[k][0]:= fi/N;
    Obj[k][2]:= res;
    Obj[k][1]:= ff/N;

    if ((res = OPT_OK) or (res = OPT_NON_CONV)) then begin
      for i:= 0 to 2 do
        for j:= 0 to 2 do
          HomN[k][i][j]:= Hn[i][j];
    end
    else begin
      for i:= 0 to 2 do
        for j:= 0 to 2 do
          HomN[k][i][j]:= 0;
    end;

    { estimation robuste }
    DimMatrix(ptsin1, 2, N-1);
    DimMatrix(ptsin2, 2, N-1);
    res:= ransacfithomography(pts1, pts2, ptsin1, ptsin2, H, tolransac, ninliers);
    Obj[k][3]:= ninliers;
    Obj[k][4]:= res;
    for i:= 0 to 2 do
      for j:= 0 to 2 do
        HomR[k][i][j]:= H[i][j];

    { estimation non linaire aprs robuste}
    //DimMatrix(Hn, 2, 2);

    for i:= 0 to 2 do begin
      SetLength(ptsin1[i], ninliers); // modif 24/05/04
      SetLength(ptsin2[i], ninliers);
    end;
    res:= homography2dnlin(ptsin1, ptsin2, H, Hn,maxiter,fi, ff);
    Obj[k][5]:= fi/ninliers;
    Obj[k][7]:= res;
    Obj[k][6]:= ff/ninliers;

    if ((res = OPT_OK) or (res = OPT_NON_CONV)) then begin
      for i:= 0 to 2 do
        for j:= 0 to 2 do
          HomRN[k][i][j]:= Hn[i][j];
    end
    else begin
      for i:= 0 to 2 do
        for j:= 0 to 2 do
          HomRN[k][i][j]:= 0;
    end;

    // sauver les points pour affichage
    SetLength(PtsIntAppRob[k], ninliers);
    for i:= 0 to ninliers-1 do begin
      PtsIntAppRob[k][i].i:= ptsin1[0][i];
      PtsIntAppRob[k][i].j:= ptsin1[1][i];
      PtsIntAppRob[k][i].u:= ptsin2[0][i];
      PtsIntAppRob[k][i].v:= ptsin2[1][i];
    end;
  end;
  PtsIntAppRobDrawable:= true;
  ImageG.PaintBuffer;
  ImageG.RefreshBox;
end;

{
procedure EstimeH(MaxIter: Integer);
var
  k, l, N, M, p, i, j: Integer;
  NewPts: array of array of TRPoint;
  mi, mj, mu, mv, scale1, scale2, x, y, w: Real;
  A, V, Xx: TMatrix;
  S: TVector;
  H, T1, T2, Ht, Ht2: array[0..2,0..2] of Real;
  XV: TVector;
  F_min: Float;
  H_inv: TMatrix;
  res: Integer;
begin
  SetLength(NewPts, Length(PtsIntApp));
  SetLength(Hom, Length(PtsIntApp));
  SetLength(HomN, Length(PtsIntApp));
  SetLength(Obj, Length(PtsIntApp));
  DimVector(XV, 9);
  DimMatrix(H_inv, 8, 8);
  for k:= 0 to Length(PtsIntApp)-1 do begin
    SetLength(NewPts[k], Length(PtsIntApp[k]));
    mi:= 0;
    mj:= 0;
    mu:= 0;
    mv:= 0;
    N:= Length(NewPts[k]);
    M:= 9;
    for l:= 0 to N-1 do begin
      mi:= mi + PtsIntApp[k][l].i;
      mj:= mj + PtsIntApp[k][l].j;
      mu:= mu + PtsIntApp[k][l].u;
      mv:= mv + PtsIntApp[k][l].v;
    end;
    mi:= mi / N;
    mj:= mj / N;
    mu:= mu / N;
    mv:= mv / N;
    for l:= 0 to N-1 do begin
      NewPts[k][l].i:= PtsIntApp[k][l].i - mi;
      NewPts[k][l].j:= PtsIntApp[k][l].j - mj;
      NewPts[k][l].u:= PtsIntApp[k][l].u - mu;
      NewPts[k][l].v:= PtsIntApp[k][l].v - mv;
    end;

    scale1:= 0;
    scale2:= 0;
    for l:= 0 to N-1 do begin
      scale1:= scale1 + sqrt( NewPts[k][l].i*NewPts[k][l].i + NewPts[k][l].j*NewPts[k][l].j );
      scale2:= scale2 + sqrt( NewPts[k][l].u*NewPts[k][l].u + NewPts[k][l].v*NewPts[k][l].v );
    end;
    scale1:= scale1 / N;
    scale2:= scale2 / N;
    scale1:= sqrt(2)/scale1;
    scale2:= sqrt(2)/scale2;

    for l:= 0 to N-1 do begin
      x:= PtsIntApp[k][l].i;
      y:= PtsIntApp[k][l].j;
      NewPts[k][l].i:= scale1*x - scale1*mi;
      NewPts[k][l].j:= scale1*y - scale1*mj;
      x:= PtsIntApp[k][l].u;
      y:= PtsIntApp[k][l].v;
      NewPts[k][l].u:= scale2*x - scale2*mu;
      NewPts[k][l].v:= scale2*y - scale2*mv;
    end;

    DimMatrix(A, 3*N, M);
    for l:= 0 to N-1 do begin
      x:= NewPts[k][l].u;
      y:= NewPts[k][l].v;
      w:= 1;
      A[3*l][0]:= 0;
      A[3*l][1]:= 0;
      A[3*l][2]:= 0;
      A[3*l][3]:= -w*NewPts[k][l].i;
      A[3*l][4]:= -w*NewPts[k][l].j;
      A[3*l][5]:= -w*1;
      A[3*l][6]:= y*NewPts[k][l].i;
      A[3*l][7]:= y*NewPts[k][l].j;
      A[3*l][8]:= y*1;

      A[3*l+1][0]:= w*NewPts[k][l].i;
      A[3*l+1][1]:= w*NewPts[k][l].j;
      A[3*l+1][2]:= w*1;
      A[3*l+1][3]:= 0;
      A[3*l+1][4]:= 0;
      A[3*l+1][5]:= 0;
      A[3*l+1][6]:= -x*NewPts[k][l].i;
      A[3*l+1][7]:= -x*NewPts[k][l].j;
      A[3*l+1][8]:= -x*1;

      A[3*l+2][0]:= -y*NewPts[k][l].i;
      A[3*l+2][1]:= -y*NewPts[k][l].j;
      A[3*l+2][2]:= -y*1;
      A[3*l+2][3]:= x*NewPts[k][l].i;
      A[3*l+2][4]:= x*NewPts[k][l].j;
      A[3*l+2][5]:= x*1;
      A[3*l+2][6]:= 0;
      A[3*l+2][7]:= 0;
      A[3*l+2][8]:= 0;
    end;
    DimVector(S, M);
    DimMatrix(V, M, M);

    if SV_Decomp(A, 0, 3*N-1, M-1, S, V) <> MAT_OK then begin
      HomN[k][0][0]:= 0;
      HomN[k][1][0]:= 0;
      HomN[k][2][0]:= 0;
      HomN[k][0][1]:= 0;
      HomN[k][1][1]:= 0;
      HomN[k][2][1]:= 0;
      HomN[k][0][2]:= 0;
      HomN[k][1][2]:= 0;
      HomN[k][2][2]:= 0;
      Hom[k][0][0]:= 0;
      Hom[k][1][0]:= 0;
      Hom[k][2][0]:= 0;
      Hom[k][0][1]:= 0;
      Hom[k][1][1]:= 0;
      Hom[k][2][1]:= 0;
      Hom[k][0][2]:= 0;
      Hom[k][1][2]:= 0;
      Hom[k][2][2]:= 0;
      Obj[k][0]:= 0;
      Obj[k][2]:= MAT_NON_CONV;
      Obj[k][1]:= 0;
      Continue;
    end;

    SV_SetZero(S, 0, M-1, TOL);

    w:= S[0];
    p:= 0;
    for l:= 0 to M-1 do begin
      if (S[l] <= w) then begin
        w:= S[l];
        p:= l;
      end;
    end;
    for l:= 0 to M-1 do
      Ht[Floor(l/3)][l mod 3]:= V[l][p];
    T1[0][0]:= scale1;
    T1[0][1]:= 0;
    T1[0][2]:= -scale1*mi;
    T1[1][0]:= 0;
    T1[1][1]:= scale1;
    T1[1][2]:= -scale1*mj;
    T1[2][0]:= 0;
    T1[2][1]:= 0;
    T1[2][2]:= 1;

    T2[0][0]:= 1/scale2;
    T2[0][1]:= 0;
    T2[0][2]:= mu;
    T2[1][0]:= 0;
    T2[1][1]:= 1/scale2;
    T2[1][2]:= mv;
    T2[2][0]:= 0;
    T2[2][1]:= 0;
    T2[2][2]:= 1;

    for i:= 0 to 2 do begin
      for j:= 0 to 2 do begin
        Ht2[i][j]:= T2[i][0]*Ht[0][j]+T2[i][1]*Ht[1][j]+T2[i][2]*Ht[2][j];
      end;
    end;
    for i:= 0 to 2 do begin
      for j:= 0 to 2 do begin
        Hom[k][i][j]:= Ht2[i][0]*T1[0][j]+Ht2[i][1]*T1[1][j]+Ht2[i][2]*T1[2][j];
      end;
    end;
    XV[0]:= Hom[k][0][0];
    XV[1]:= Hom[k][1][0];
    XV[2]:= Hom[k][2][0];
    XV[3]:= Hom[k][0][1];
    XV[4]:= Hom[k][1][1];
    XV[5]:= Hom[k][2][1];
    XV[6]:= Hom[k][0][2];
    XV[7]:= Hom[k][1][2];
    XV[8]:= Hom[k][2][2];
    XV[9]:= k;
    KK:= k;
    Obj[k][0]:= FObj(XV);
    res:= Marquardt(FObj, NumHessGrad, XV, 0, 8, 20, 1E-10, F_min, H_inv);
    Obj[k][2]:= res;
    Obj[k][1]:= F_min;
    if ((res = OPT_OK) or (res = OPT_NON_CONV)) then begin
    HomN[k][0][0]:= XV[0];
    HomN[k][1][0]:= XV[1];
    HomN[k][2][0]:= XV[2];
    HomN[k][0][1]:= XV[3];
    HomN[k][1][1]:= XV[4];
    HomN[k][2][1]:= XV[5];
    HomN[k][0][2]:= XV[6];
    HomN[k][1][2]:= XV[7];
    HomN[k][2][2]:= XV[8];
    end
    else begin
    HomN[k][0][0]:= Hom[k][0][0];
    HomN[k][1][0]:= Hom[k][1][0];
    HomN[k][2][0]:= Hom[k][2][0];
    HomN[k][0][1]:= Hom[k][0][1];
    HomN[k][1][1]:= Hom[k][1][1];
    HomN[k][2][1]:= Hom[k][2][1];
    HomN[k][0][2]:= Hom[k][0][2];
    HomN[k][1][2]:= Hom[k][1][2];
    HomN[k][2][2]:= Hom[k][2][2];
    end;
  end;
end;
}

function CCNC(f1, f2: ATRPoint): Real;
var
  n, i: Integer;
  sum1, sum2, norm1, norm2, mean1, mean2: Real;
begin
  n:= Length(f1);
  sum1:= 0;
  sum2:= 0;
  for i:= 0 to n-1 do begin
    sum1:= sum1 + f1[i].val;
    sum2:= sum2 + f2[i].val;
  end;
  mean1:= sum1 / n;
  mean2:= sum2 / n;
  norm1:= 0;
  norm2:= 0;
  for i:= 0 to n-1 do begin
    f1[i].val:= f1[i].val - mean1;
    f2[i].val:= f2[i].val - mean2;
    norm1:= norm1 + f1[i].val*f1[i].val;
    norm2:= norm2 + f2[i].val*f2[i].val;
  end;
  norm1:= sqrt(norm1);
  norm2:= sqrt(norm2);
  if (norm1 = 0) then
    norm1:= 1e-5;
  if (norm2 = 0) then
    norm2:= 1e-5;

  Result:= 0;
  for i:= 0 to n-1 do begin
    Result:= Result + (f1[i].val*f2[i].val);
  end;
  Result:= Result /(norm1*norm2);
end;

function Corresp(Z1,Z2: TRect; ImDep, ImArr: TBitmap; polyArr: TList): TRPoint;
var
  m, n, n2, nblig, nbcol, h, c, n1, nbl, nbc: Integer;
  invnn, v: Real;
  res: TRPoint;
  rowsg, rowsd: array of pByteArray;
  i, j, ind, ind2: Integer;
  val, f1, f2: ATRPoint;
  ok: Boolean;
begin
  m:= Z1.Bottom-Z1.Top+1;
  n:= Z1.Right-Z1.Left+1;
  nblig:= Z2.Bottom-Z2.Top+1;
  nbcol:= Z2.Right-Z2.Left+1;
  SetLength(rowsg, m);
  for h:= 0 to m-1 do
    rowsg[h]:= ImDep.ScanLine[Z1.Top+h];

  SetLength(f1, n*n);
  SetLength(f2, n*n);
  ind:= 0;
  for h:= 0 to m-1 do begin
    for c:= 0 to n-1 do begin
      f1[ind].val:= rowsg[h][Z1.Left+c];
      f1[ind].i:= Z1.Top+h;
      f1[ind].j:= Z1.Left+c;
      Inc(ind);
    end;
  end;

  n1:= n-1;
  n2:= Round(n1/2);
  invnn:= 1/(n*n);
  nbl:= nblig-n1;
  nbc:= nbcol-n1;
  SetLength(val, nbl*nbc);

  SetLength(rowsd, nblig);
  for h:= 0 to nblig-1 do
    rowsd[h]:= ImArr.ScanLine[Z2.Top+h];

  ind2:= 0;
  for i:= 0 to nbl-1 do begin
    for j:= 0 to nbc-1 do begin
      ind:= 0;
      val[ind2].U:= Z2.Top+i+n2;
      val[ind2].V:= Z2.Left+j+n2;
      for h:= i to i+n1 do begin
        for c:= j to j+n1 do begin
          f2[ind].U:= Z2.Top+h;
          f2[ind].V:= Z2.Left+c;
          f2[ind].val:= rowsd[h][Z2.Left+c];
          Inc(ind);
        end;
      end;
      v:= CCNC(f1, f2);
      val[ind2].val:= v;

      Inc(ind2);
    end;
  end;

  { on garde la meilleure ccnc avec la contrainte : la fentre de corrlation
    pour ce point doit tre incluse entirement dans le polygone }
  Result.val:= -1;
  while (true) do begin
    res:= MaxTab(val);
    if (res.val = -1) then Break;
    ok:= true;
    ind:= FindPolygon(polyArr, res.U, res.V);
    for i:= -n2 to n2 do begin
      for j:= -n2 to n2 do begin
        ind2:= FindPolygon(polyArr, res.U+i, res.V+j);
        if (ind2 <> ind) then begin
          ok:= false;
          Break;
        end;
      end;
      if (not ok) then Break;
    end;
    if (ok) then begin
      Result:= res;
      Break;
    end
    else begin
      val[res.ind].val:= -1;
    end;
  end;
end;

function SubCorresp(Z1,Z2: TRect; prec: Real; bidir, p1, p2, q1, q2: Integer): TRPoint;
var
  m, n, n2, nblig, nbcol: Integer;
  invnn, shift, cc, cccur: Real;
  res, res2: TRPoint;
  i, j, ish, jsh, k, l, ind,h,c, pol: Integer;
  f1, f2: ATRPoint;
  mat: TMatInt;
  row: pByteArray;
  subpixlig, subpixcol: Real;
  rowsg: array of pByteArray;
  Z3, Z4: TRect;
  a1,a2,a3,a4: Integer;
begin
  m:= Z1.Bottom-Z1.Top+1;
  n:= Z1.Right-Z1.Left+1;
  n2:= Round((n-1)/2);
  nblig:= Z2.Bottom-Z2.Top+1;
  nbcol:= Z2.Right-Z2.Left+1;
  invnn:= 1/(n*n);

  { recherche du meilleur correspondant entier }
  res:= Corresp(Z1,Z2,ImageG.TheImage,ImageD.TheImage,ImageD.Polygones);

  { contrainte bidirectionnelle souple }
  if (bidir <> -1) then begin
    i:= Round(res.U);
    j:= Round(res.V);
    Z3.Left:= j - n2;
    Z3.Right:= j + n2;
    Z3.Top:= i - n2;
    Z3.Bottom:= i + n2;

    a1:= math.Max(j-q2-n2,0);
    a2:= math.Max(i-p2-n2,0);
    a3:= math.Min(j+q1+n2,ImageG.TheImage.Width-1);
    a4:= math.Min(i+p2+n2,ImageG.TheImage.Height-1);
    Z4:= Rect(a1, a2, a3, a4);

    res2:= Corresp(Z3,Z4,ImageD.TheImage,ImageG.TheImage,ImageG.Polygones);
    if ( (abs(res2.V - (Z1.Left+n2)) > bidir) or (abs(res2.U - (Z1.Top+n2)) > bidir) ) then
      res.val:= -1;
  end;

  { si on l'a trouv alors recherche au sous-pixel
    condition que le pixel ne soit pas trop prs des bords
   pour l'interpolation bicubique, auquel cas on l'invalide }
  if ((res.val <> -1) and (prec <> 1)) then begin
    subpixlig:= Round(res.U);
    subpixcol:= Round(res.V);
    if ((subpixlig <= (n2+1)) or (subpixlig >= (ImageD.TheImage.Height-n2-4))
      or (subpixcol <= (n2+1)) or (subpixcol >= (ImageD.TheImage.Width-n2-4))) then begin
      Result.val:= -1;
      Exit;
    end;
    pol:= FindPolygon(ImageD.Polygones, res.U, res.V);

    { copie de l'image droite dans un pMatInt pour les interpolations }
    setLength(mat, ImageD.TheImage.Height);
    for i:= 0 to Length(mat)-1 do
      setLength(mat[i], ImageD.TheImage.Width);
    for i:= 0 to Length(mat)-1 do begin
      row:= ImageD.TheImage.ScanLine[i];
      for j:= 0 to Length(mat[0])-1 do begin
        mat[i][j]:= row[j];
      end;
    end;

    { init des tableaux }
    SetLength(f1, n*n);
    SetLength(f2, n*n);
    SetLength(rowsg, m);
    for h:= 0 to m-1 do
      rowsg[h]:= ImageG.TheImage.ScanLine[Z1.Top+h];
    ind:= 0;
    for h:= 0 to m-1 do begin
      for c:= 0 to n-1 do begin
        f1[ind].val:= rowsg[h][Z1.Left+c];
        f1[ind].i:= Z1.Top+h;
        f1[ind].j:= Z1.Left+c;
        Inc(ind);
      end;
    end;

    { distance de dcalage pour la recherche du sous-pixel }
    shift:= 1;

    { facteurs de dcalages (le sous pixel est dcal de (ish*shift,jsh*shift) }
    ish:= 0;
    jsh:= 0;

    { recherche itrative du meilleur sous-pixel jusqu' atteindre la prcision voulue }
    cc:= -1;
    while (shift > prec) do begin
      shift:= shift/2;
      cc:= -1E10;

      { recherche du meilleur correspondant parmi les 8 sous-pixels voisins
        et le sous pixel lui-meme }
      for i:= -1 to 1 do begin
        for j:= -1 to 1 do begin
          { calcul des niveaux de gris des sous-pixels par interpolation }
          ind:= 0;
          for k:= -n2 to n2 do begin
            for l:= -n2 to n2 do begin
              f2[ind].U:= subpixlig+i*shift+k;
              f2[ind].V:= subpixcol+j*shift+l;
              f2[ind].val:= interpBicubique(f2[ind].U, f2[ind].V, @mat);
              Inc(ind);
            end;
          end;
          cccur:= CCNC(f1, f2);
          if (cccur > cc) then begin
            ish:= i;
            jsh:= j;
            cc:= cccur;
          end;
        end;
      end;
      { dcalage du sous-pixel vers le meilleur voisin trouv }
      subpixcol:= subpixcol+jsh*shift;
      subpixlig:= subpixlig+ish*shift;
    end;
    res.U:= subpixlig;
    res.V:= subpixcol;
    res.val:= cc;

    { si le sous pixel n'est pas dans mme polygone que le pixel
      on l'invalide }
    if (FindPolygon(ImageD.Polygones, res.U, res.V) <> pol) then
      res.val:= -1;
  end;
  Result:= res;
end;

procedure Apparie(n,p1,p2,q1,q2: Integer; ccncMin: Real; prec: Real; bidir: Integer);
var
  k, l, n2, i, j, p, i2, j2, ind: Integer;
  ok: Boolean;
  hglig, hgcol, bdlig, bdcol: Integer;
  Z1, Z2: TRect;
  res: TRPoint;
begin
  n2:= Round((n-1)/2);
  SetLength(PtsIntApp, Length(PtsInt));
  SetLength(PtsIndexApp, Length(PtsInt));
  { boucles sur les points d'intrt trouvs }
  for k:= 0 to Length(PtsIntApp)-1 do begin
    if (Polygone(ImageG.Polygones[k]).Locked) then
      Continue;
    SetLength(PtsIntApp[k], Length(PtsInt[k]));
    PtsIndexApp[k]:= 0;
    for l:= 0 to Length(PtsInt[k])-1 do begin
      i:= Round(PtsInt[k][l].U);
      j:= Round(PtsInt[k][l].V);

      //ok:= true;

      {
      p:= FindPolygon(ImageG.Polygones, n2, j);
      for i2:= 0 to n-1 do begin
        for j2:= -n2 to n2 do begin
          if (FindPolygon(ImageG.Polygones, i2, j+j2) <> p) then begin
            ok:= false;
            break;
          end;
        end;
      end;
      }

      //if (ok) then begin
        Z1:= Rect(j-n2, i-n2, j+n2, i+n2);
        hglig:= math.Max(i-p1-n2,0);
        hgcol:= math.Max(j-q1-n2,0);
        bdlig:= math.Min(i+p1+n2,ImageD.TheImage.Height-1);
        bdcol:= math.Min(j+q2+n2,ImageD.TheImage.Width-1);
        Z2:= Rect(hgcol, hglig, bdcol, bdlig);
        res:= SubCorresp(Z1,Z2,prec,bidir,p1,p2,q1,q2);

        if (res.val >= ccncMin) then begin
          res.i:= i;
          res.j:= j;
          PtsIntApp[k][PtsIndexApp[k]]:= res;
          Inc(PtsIndexApp[k]);
        end;
      //end;
    end;
  end;
  { rajustement des tailles }
  for k:= 0 to Length(PtsIntApp)-1 do begin
    SetLength(PtsIntApp[k], PtsIndexApp[k]);
  end;
  PtsIntAppDrawable:= true;
  PtsIntAppRobDrawable:= false;
  ImageG.PaintBuffer;
  ImageG.RefreshBox;
  ImageD.PaintBuffer;
  ImageD.RefreshBox;
end;

function MaxLocalTab(tab: array of TRPoint; U, V: Real; n2: Integer): TRPoint;
var
  i: Integer;
begin
  Result.val:= -INTEMAX;
  for i:= 0 to Length(tab)-1 do begin
    if ((tab[i].U < (U-n2)) or (tab[i].U > (U+n2)) or (tab[i].V < (V-n2)) or (tab[i].V > (V+n2))) then continue;
    if (tab[i].val > Result.val) then begin
      Result.val:= tab[i].val;
      Result.ind:= i;
    end;
  end;
end;

function MinTab(tab: array of TRPoint): TRPoint;
var
  i: Integer;
begin
  Result.val:= INTEMAX;
  for i:= 0 to Length(tab)-1 do begin
    if (tab[i].val < Result.val) then begin
      Result.val:= tab[i].val;
      Result.ind:= i;
    end;
  end;
end;

function MaxTab(tab: array of TRPoint): TRPoint;
var
  i: Integer;
begin
  Result.val:= -INTEMAX;
  for i:= 0 to Length(tab)-1 do begin
    if (tab[i].val > Result.val) then begin
      Result:= tab[i];
      //Result.val:= tab[i].val;
      Result.ind:= i;
    end;
  end;
end;

procedure DetectPts(polygones: TList; TheImage: TBitmap; N, N_poly: Integer; Npoints: Integer);
var
  i, j, h, n2, n2_poly, i2, j2, NLig, NCol: Integer;
  p: Integer;
  rows: array of pByteArray;
  ok: Boolean;
  Ii, Ij, sigma, lambda, a, b, c: Real;
  gauss, gaussd: array of Real;
  gauss2d: array of array of Real;
  tabmin: TRPoint;
  poly: Polygone;
  bord: Integer;
  reponse, Iis, Ijs: array of array of Real;
  Itmp, s1, s2: Real;
begin

  n2:= Round((N-1)/2);
  n2_poly:= Round((N_poly-1)/2);
  SetLength(rows, N);
  NLig:= TheImage.Height;
  NCol:= TheImage.Width;
  SetLength(reponse, TheImage.Height);
  for i:= 0 to TheImage.Height-1 do
    SetLength(reponse[i], TheImage.Width);

  { on mmorise  quel polygone appartiennent les pixels
    entiers de l'image de gauche }
  SetLength(PolyMap, NLig);
  for i:= 0 to NLig-1 do begin
    SetLength(PolyMap[i], NCol);
    for j:= 0 to NCol-1 do begin
      PolyMap[i][j]:= FindPolygon(Polygones, i, j);
    end;
  end;

  { stockage de la carte des polygones si besoin }
  

  { init du stockage des points d'intrt }
  SetLength(PtsInt, Polygones.Count);
  SetLength(PtsIndex, Polygones.Count);
  for i:= 0 to Length(PtsInt)-1 do begin
    if (Polygone(Polygones[i]).Locked) then Continue;
    SetLength(PtsInt[i], Npoints);
    for j:= 0 to Npoints-1 do
      PtsInt[i][j].val:= -100;
    PtsIndex[i]:= 0;
  end;

  { initialisation des masques }
  sigma:= N/4;
  SetLength(gaussd, N);
  SetLength(gauss, N);
  SetLength(gauss2d, N);
  for i:= 0 to N-1 do
    SetLength(gauss2d[i], N);
  s1:= 0;
  s2:= 0;
  for i:= -n2 to n2 do begin
    gauss[i+n2]:= 1/(sqrt(2*PI)*sigma)*exp(-i*i/(2*sigma*sigma));
    s1:= s1 + gauss[i+n2];
    gaussd[i+n2]:= -i/(sqrt(2*PI)*sigma*sigma*sigma)*exp(-i*i/(2*sigma*sigma));
    s2:= s2 - i*gaussd[i+n2];
    //s2:= s2 -sign(i)* gaussd[i+n2];
  end;
  { normalisation des masques }
  for i:= -n2 to n2 do begin
    gauss[i+n2]:= gauss[i+n2] / s1;
    gaussd[i+n2]:= gaussd[i+n2] / s2;
  end;

  s1:= 0;
  for i:= -n2 to n2 do begin
    for i2:= -n2 to n2 do begin
      gauss2d[i+n2][i2+n2]:= 1/(2*PI*sigma*sigma)*exp(-(i*i+i2*i2)/(2*sigma*sigma));
      s1:= s1 + gauss2d[i+n2][i2+n2];
    end;
  end;

  for i:= -n2 to n2 do begin
    for i2:= -n2 to n2 do begin
      gauss2d[i+n2][i2+n2]:= gauss2d[i+n2][i2+n2] / s1;
    end;
  end;

  { boucle sur l'image pour calculer les gradients }
  bord:= math.max(n2, n2_poly);
  SetLength(Iis, NLig);
  SetLength(Ijs, Nlig);
  for h:= 0 to NLig-1 do begin
    SetLength(Iis[h], Ncol);
    SetLength(Ijs[h], NCol);

    if ((h < bord) or (h > (NLig-bord-1))) then begin
      for j:= 0 to NCol-1 do begin
        Iis[h][j]:= 0;
        Ijs[h][j]:= 0;
      end;
      Continue;
    end;

    { rcuprer les lignes dans l'image gauche }
    for i:= 0 to N-1 do begin
      rows[i]:= TheImage.ScanLine[h-n2+i];
    end;

    for j:= 0 to NCol-1 do begin
      if ((j < bord) or (j > (NCol-bord-1))) then begin
        Iis[h][j]:= 0;
        Ijs[h][j]:= 0;
        Continue;
      end;


      if ((PolyMap[h][j] = -1) or (Polygone(Polygones[PolyMap[h][j]]).Locked)) then begin
        Iis[h][j]:= -INTEMAX;
        Ijs[h][j]:= -INTEMAX;
      end
      else begin
        { lissage horizontal et drive + lissage vertical }
        Iis[h][j]:= 0;
        for i2:= -n2 to n2 do begin
          Itmp:= 0;
          for i:= 0 to N-1 do begin
            Itmp:= Itmp + rows[i][j+i2]*gauss[N-1-i];
          end;
          Iis[h][j]:= Iis[h][j] + Itmp*gaussd[N-1-(i2+n2)];
        end;
        { lissage vertical et drive + lissage horizontal }
        Ijs[h][j]:= 0;
        for i:= 0 to N-1 do begin
          Itmp:= 0;
          for i2:= -n2 to n2 do begin
            Itmp:= Itmp + rows[i][j+i2]*gauss[N-1-(i2+n2)];
          end;
          Ijs[h][j]:= Ijs[h][j] + Itmp*gaussd[N-1-i];
        end;
      end;
    end;
  end;

  for h:= bord to NLig-1-bord do begin

    { on effectue le travail sur une ligne, rows[n2] }
    for j:= bord to NCol-1-bord do begin
      { tester si le point est assez loin d'une frontire de polygone }
      ok:= true;

      { on cherche  savoir si la fentre N_polyxN_poly est incluse
        dans le mme polygone }
      //p:= FindPolygon(Polygones, h, j);
      p:= PolyMap[h][j];
      { polygone bloqu }
      if (p <> -1) then
        if (Polygone(Polygones[p]).Locked) then begin
          reponse[h][j]:= -INTEMAX;
          Continue;
        end;

      if (p <> -1) then begin
        { on teste seulement le primtre de la fentre pour acclrer }
        for i2:= 0 to N_poly-1 do begin
          //if (FindPolygon(Polygones, h-n2_poly+i2, j-n2_poly) <> p) then begin
          if (PolyMap[h-n2_poly+i2][j-n2_poly] <> p) then begin
            ok:= false;
            break;
          end;
          //if (FindPolygon(Polygones, h-n2_poly+i2, j+n2_poly) <> p) then begin
          if (PolyMap[h-n2_poly+i2][j+n2_poly] <> p) then begin
            ok:= false;
            break;
          end;
        end;
        if (ok) then begin
          for j2:= -n2_poly to n2_poly do begin
            //if (FindPolygon(Polygones, h-n2_poly, j+j2) <> p) then begin
            if (PolyMap[h-n2_poly][j+j2] <> p) then begin
              ok:= false;
              break;
            end;
            //if (FindPolygon(Polygones, h+n2_poly, j+j2) <> p) then begin
            if (PolyMap[h+n2_poly][j+j2] <> p) then begin
              ok:= false;
              break;
            end;
          end;
        end;

        if (ok) then begin
          a:= 0;
          b:= 0;
          c:= 0;
          for i:= -n2 to n2 do begin
            for i2:= -n2 to n2 do begin
              a:= a + gauss2d[i+n2][i2+n2]*Iis[h+i][j+i2]*Iis[h+i][j+i2];
              c:= c + gauss2d[i+n2][i2+n2]*Ijs[h+i][j+i2]*Ijs[h+i][j+i2];
              b:= b + gauss2d[i+n2][i2+n2]*Iis[h+i][j+i2]*Ijs[h+i][j+i2];
            end;
          end;

          reponse[h][j]:= 0.5*(a+c-sqrt((a-c)*(a-c)+4*b*b));
        end
        else
          reponse[h][j]:= -INTEMAX;
      end
      else
        reponse[h][j]:= -INTEMAX;
    end;
  end;

  { supression des non maxima }
  for h:= bord to NLig-1-bord do begin
    for j:= bord to NCol-1-bord do begin
      p:= PolyMap[h][j];
      if ((p = -1) or (Polygone(Polygones[p]).Locked) or (reponse[h][j] = -INTEMAX)) then
        Continue;
      { si pas de max local alors on stocke le points }
      ok:= true;
      for i:= -n2 to n2 do begin
        for i2:= -n2 to n2 do begin
          if (reponse[h+i][j+i2] > reponse[h][j]) then
            ok:= false;
        end;
      end;
      if (ok) then begin
        { stockage si tab pas rempli ou si infrieur au min du tableau }
        if (PtsIndex[p] <= (Npoints-1)) then begin
          PtsInt[p][PtsIndex[p]].val:= reponse[h][j];
          PtsInt[p][PtsIndex[p]].U:= h;
          PtsInt[p][PtsIndex[p]].V:= j;
          Inc(PtsIndex[p]);
        end
        else begin
          tabmin:= MinTab(PtsInt[p]);
          if (reponse[h][j] > tabmin.val) then begin
            PtsInt[p][tabmin.ind].val:= reponse[h][j];
            PtsInt[p][tabmin.ind].U:= h;
            PtsInt[p][tabmin.ind].V:= j;
          end;
        end;
      end;
    end;
  end;

  { ajustement des tailles }
  for i:= 0 to Length(PtsInt)-1 do begin
    SetLength(PtsInt[i], PtsIndex[i]);
  end;
end;

procedure AugmenteE_Segments(N: Integer);
var
  i, i0: Integer;
begin
  i0:= Length(E_Segments);
  SetLength(E_Segments, i0+N);
  for i:= i0 to Length(E_Segments)-1 do
    SetLength(E_Segments[i], 5);
end;

procedure VideE_Segments;
begin
  E_Segments:= nil;
  E_Segments_Index:= 0;
end;

procedure InitE_Segments(N: Integer);
var
  i: Integer;
begin
  VideE_Segments;
  SetLength(E_Segments, N);
  for i:= 0 to N-1 do
    SetLength(E_Segments[i], 5);
end;

procedure StockeE_Segments(u1, v1, u2, v2, E: Real);
begin
  { si toute la place est prise on agrandit }
  if (E_Segments_Index >= Length(E_Segments)) then
    AugmenteE_Segments(1000);
  E_Segments[E_Segments_Index, 0]:= u1;
  E_Segments[E_Segments_Index, 1]:= v1;
  E_Segments[E_Segments_Index, 2]:= u2;
  E_Segments[E_Segments_Index, 3]:= v2;
  E_Segments[E_Segments_Index, 4]:= E;
  Inc(E_Segments_Index);
end;

function ChercheE_Segments(u1, v1, u2, v2: Real): Real;
var
  i: Integer;
begin
  for i:= 0 to E_Segments_Index-1 do begin
    if ((E_Segments[i, 0]=u1) and (E_Segments[i, 1]=v1) and (E_Segments[i, 2]=u2) and (E_Segments[i, 3]=v2)) then begin
      Result:= E_Segments[i, 4];
      Exit;
    end;
  end;
  Result:= -100;
end;

function polySnake2(imgradx, imgrady: pMatInt; polyf: Polygone; pas: Real; nite: Integer; profondeur: array of Integer; nsom, maxiter: Integer): Polygone;
var
  k, i, l: Integer;
  ite: Integer;
  cvg: Boolean;
  Emax: Real;
  s: Sommet;
  taille: Real;
  polyi, polys: Polygone;
  pasadapt: Boolean;
  ss: string;

  N: Integer;
  X, Xmin, Xmax: TVector;
  ErrCode: Integer;
  F_min: Float;
  H_inv: TMatrix;
  nsomc: Integer;
begin
  if (pas = 0) then
    pasadapt:= true
  else
    pasadapt:= false;

  { le polygone final est cree une fois pour toute
    il est initialise au poly init (les refs des sommets)
    ensuite c'est juste les U et V des sommets qui varieront  }
  //polyf:= Polygone.Create;
  polys:= Polygone.Create;
  polyEval:= Polygone.Create;
  //ConcatListe(polyf.Sommets, polys.Sommets);
  //ConcatListe(polyEval.Sommets, polys.Sommets);

  { le polys va etre modifie alors il lui faut des nouvelles refs
    idem pour le polyEval si mthode incomplte }
  //polys.Sommets.Clear;
  polys.Clone(polyf);
  //polyEval.Sommets.Clear;
  polyEval.Clone(polyf);

  for k:= 0 to Length(profondeur)-1 do begin
    FRapport.Write('              Profondeur niveau '+IntToStr(k+1)+' : prcision 1/'+IntToStr(profondeur[k]));
    FRapport.Write('              ---------------------------------------------------------------');
    FRapport.Write(' ');
    ite:= 1;
    cvg:= false;
    Emax:= -100;

    { ajuster les sommets sur les bords }
    { et compter les sommets non coins et non bloqus }
    nsomc:= 0;
    for i:= 0 to polys.Sommets.Count-1 do begin
      s:= Sommet(polys.Sommets[i]);
      if ((s.BordU) and (s.U <> 0)) then begin
        s.U:= Floor(s.U) + (1 - 1/profondeur[k]);
      end;
      if ((s.BordV) and (s.V <> 0)) then
        s.V:= Floor(s.V) + (1 - 1/profondeur[k]);
      if ((i <> 0) and (not(s.BordU and s.BordV)) and (not s.Bloque)) then
        Inc(nsomc);
    end;

    taille:= 1/profondeur[k];
    if (pasadapt) then
      pas:= taille;
    InitE_Segments(81*polys.Sommets.Count-1);

    while ((ite <= nite) and not cvg) do begin
      FRapport.Write('                  * Itration '+IntToStr(ite)+' : prcision 1/'+IntToStr(profondeur[k]));
      ss:= '                      - Polygone initial : ';
      for l:= 0 to polys.Sommets.Count-1 do
        ss:= ss + '('+Format('%5.1f', [Sommet(polys.Sommets[l]).U]) + ' ' + Format('%5.1f', [Sommet(polys.Sommets[l]).V]) + ') ';
      FRapport.Write(ss);

      if (ite <> 1) then
        AugmenteE_Segments(81*polys.Sommets.Count-1);

      { choix de la mthode de recherche de la solution }   
      if ((maxiter = 0) or ((maxiter <> 0) and (nsom <> 0) and (nsomc < nsom))) then begin
        { methode complte }
        polyi:= Polygone.Create;
        polyf.Energie:= 0;
        ParcoursEspace(0, polys.Sommets.Count-2, imgradx, imgrady, polys, polyi, taille, pas, -100, polyf);
        polyi.Free;
      end
      else begin
        { methode incomplte }
        N:= polys.Sommets.Count-1;
        DimVector(X, 2*N-1);
        DimVector(Xmin, 2*N-1);
        DimVector(Xmax, 2*N-1);
        DimMatrix(H_inv, 2*N-1, 2*N-1);
        for l:= 0 to N-1 do begin
          s:= Sommet(polys.Sommets[l]);
          X[2*l]:= s.U;
          X[2*l+1]:= s.V;
          Xmin[2*l]:= X[2*l] - taille;
          Xmax[2*l]:= X[2*l] + taille;
          Xmin[2*l+1]:= X[2*l+1] - taille;
          Xmax[2*l+1]:= X[2*l+1] + taille;
          if (s.BordU) then begin
            Xmin[2*l]:= X[2*l];
            Xmax[2*l]:= X[2*l];
          end;
          if (s.BordV) then begin
            Xmin[2*l+1]:= X[2*l+1];
            Xmax[2*l+1]:= X[2*l+1];
          end;
          if (s.Bloque) then begin
            Xmin[2*l]:= X[2*l];
            Xmax[2*l]:= X[2*l];
            Xmin[2*l+1]:= X[2*l+1];
            Xmax[2*l+1]:= X[2*l+1];
          end;
        end;
        { les variables globales }
        pasEval:= pas;
        gradimxEval:= imgradx;
        gradimyEval:= imgrady;

        { recuit simul }
        ErrCode:= SimAnn(FObjPoly, X, Xmin, Xmax, 0, 2*N-1, 1000, taille, F_min);
        if (ErrCode <> OPT_OK) then
                FRapport.Write('                      *** Pas de convergence *** ')
        else
                FRapport.Write('                      *** Convergence *** ');


        { ajustement du minimum avec BFGS }
//      ErrCode := BFGS(FObjPoly, NumGradient, X, 0, 2*N-1, 1000, 1.0E-10, F_min, H_inv);

        { polygone final }
        for l:= 0 to N-1 do begin
          Sommet(polyf.Sommets[l]).U:= Round(X[2*l]*10)/10;
          Sommet(polyf.Sommets[l]).V:= Round(X[2*l+1]*10)/10;
        end;
        polyf.Energie:= 1/F_min;
      end;

      ss:= '                      - Polygone final :    ';
      for l:= 0 to polys.Sommets.Count-1 do
        ss:= ss + '('+Format('%5.1f', [Sommet(polyf.Sommets[l]).U]) + ' ' + Format('%5.1f', [Sommet(polyf.Sommets[l]).V]) + ') ';
      FRapport.Write(ss);
      FRapport.Write('                      - Energie finale : E = '+ Format('%10.2f',[polyf.Energie]));
      FRapport.Write(' ');
      FRapport.Memo1.Repaint;

      if (polyf.Energie > Emax) then begin
        Emax:= polyf.Energie;
        //polys.Sommets.Clear;
        polys.Clone(polyf);
      end
      else begin
        cvg:= true;
        FRapport.Write('                     *** Convergence aprs '+IntToStr(ite)+' itrations');
        FRapport.Write(' ');
      end;
      Inc(ite);
    end;
    if (ite > nite) then begin
      FRapport.Write('                     *** Pas de convergence aprs '+IntToStr(ite)+' itrations');
      FRapport.Write(' ');
      break;
    end;
  end;
  VideE_Segments;
  //Result:= polyf;
  polyEval.Free;
end;

procedure ParcoursEspace(niveau, niveau_max: Integer; imgradx, imgrady: pMatInt; polys, polyi: Polygone; taille: Real; pas: Real; Emax: Real; polyf: Polygone);
var
  Soms, somtmp: Sommet;
  polyc, polytmp: Polygone;
  Ec: Real;
  i, j: Integer;
  ii, jj: Real;
begin
//  Result:= polyf;
  Soms:= Sommet(polys.Sommets[niveau]);

  { si un sommet est sur le bord il a un degr de libert en moins
   cas sur deux bords }
  if ((Soms.BordU and Soms.BordV) or (Soms.Bloque)) then begin
    if (niveau < niveau_max) then begin
      polytmp:= Polygone.Create;
      ConcatListe(polytmp.Sommets, polyi.Sommets);
      polytmp.Sommets.Add(Soms);
      parcoursEspace(niveau+1, niveau_max, imgradx, imgrady, polys, polytmp,taille, pas, Emax, polyf);
      //Emax:= polyf.Energie;
      polytmp.Free;
    end
    else begin
      { dernier niveau -> xc, yc est un polygone complet }
      polyc:= Polygone.Create;
      ConcatListe(polyc.Sommets, polyi.Sommets);
      polyc.Sommets.Add(Soms);
      polyc.Sommets.Add(polyi.Sommets[0]);
      Ec:= EnergiePoly(polyc, pas, imgradx, imgrady);

      if (Ec > Emax) then begin
        Emax:= Ec;
        polyf.CopyUVFrom(polyc);
        polyf.Energie:= Emax;
      end;
      polyc.Free;
    end;
  end
  else if (soms.BordU) then begin
    { cas sur un bord horizontal }
    for j:= -TAILLE_FEN to TAILLE_FEN do begin
      jj:= j*taille;
      if (soms.V+jj >= 0) and ((soms.V+jj) < NCol) then begin
        somtmp:= Sommet.Create(soms.U, soms.V+jj);
        somtmp.BordU:= true;
        if (niveau < niveau_max) then begin
          polytmp:= Polygone.Create;
          ConcatListe(polytmp.Sommets, polyi.Sommets);
          polytmp.Sommets.Add(Somtmp);
          ParcoursEspace(niveau+1, niveau_max, imgradx, imgrady, polys, polytmp, taille, pas, Emax, polyf);
          Emax:= polyf.Energie;
          polytmp.Free;
          somtmp.Free;
        end
        else begin
          { dernier niveau -> xc, yc est un polygone complet }
          polyc:= Polygone.Create;
          ConcatListe(polyc.Sommets, polyi.Sommets);
          polyc.Sommets.Add(Somtmp);
          polyc.Sommets.Add(polyi.Sommets[0]);
          Ec:= EnergiePoly(polyc, pas, imgradx, imgrady);

          if (Ec > Emax) then begin
            Emax:= Ec;
            polyf.CopyUVFrom(polyc);
            polyf.Energie:= Emax;
          end;
          polyc.Free;
          somtmp.Free;
        end;
      end;
    end;
  end
  else if (soms.BordV) then begin
    { cas sur un bord vertical }
    for i:= -TAILLE_FEN to TAILLE_FEN do begin
      ii:= i*taille;
      if (soms.U+ii >= 0) and ((soms.U+ii) < NLig) then begin
        somtmp:= Sommet.Create(soms.U+ii, soms.V);
        somtmp.BordV:= true;
        if (niveau < niveau_max) then begin
          polytmp:= Polygone.Create;
          ConcatListe(polytmp.Sommets, polyi.Sommets);
          polytmp.Sommets.Add(Somtmp);
          ParcoursEspace(niveau+1, niveau_max, imgradx, imgrady, polys, polytmp, taille, pas, Emax, polyf);
          Emax:= polyf.Energie;
          polytmp.Free;
          somtmp.Free;
        end
        else begin
          { dernier niveau -> xc, yc est un polygone complet }
          polyc:= Polygone.Create;
          ConcatListe(polyc.Sommets, polyi.Sommets);
          polyc.Sommets.Add(Somtmp);
          polyc.Sommets.Add(polyi.Sommets[0]);
          Ec:= EnergiePoly(polyc, pas, imgradx, imgrady);

          if (Ec > Emax) then begin
            Emax:= Ec;
            polyf.CopyUVFrom(polyc);
            polyf.Energie:= Emax;
          end;
          polyc.Free;
          somtmp.Free;
        end;
      end;
    end;
  end
  else begin
    { cas general }
    for i:= -TAILLE_FEN to TAILLE_FEN do begin
      for j:= -TAILLE_FEN to TAILLE_FEN do begin
        ii:= i*taille;
        jj:= j*taille;
        if ( ((soms.U+ii) >= 0) and ((soms.U+ii) < NLig) and ((soms.V+jj) >= 0) and ((soms.V+jj) < NCol)) then begin
          somtmp:= Sommet.Create(soms.U+ii, soms.V+jj);
          if (niveau < niveau_max) then begin
            polytmp:= Polygone.Create;
            ConcatListe(polytmp.Sommets, polyi.Sommets);
            polytmp.Sommets.Add(Somtmp);
            ParcoursEspace(niveau+1, niveau_max, imgradx, imgrady, polys, polytmp, taille, pas, Emax, polyf);
            Emax:= polyf.Energie;
            polytmp.Free;
            somtmp.Free;
          end
          else begin
            { dernier niveau -> xc, yc est un polygone complet }
            polyc:= Polygone.Create;
            ConcatListe(polyc.Sommets, polyi.Sommets);
            polyc.Sommets.Add(Somtmp);
            polyc.Sommets.Add(polyi.Sommets[0]);
            Ec:= EnergiePoly(polyc, pas, imgradx, imgrady);

            if (Ec > Emax) then begin
              Emax:= Ec;
              polyf.CopyUVFrom(polyc);
              polyf.Energie:= Emax;
            end;
            polyc.Free;
            somtmp.Free;
          end;
        end;
      end;
    end;
  end;
end;

function EnergiePoly(polys: Polygone; pas: Real; gradimx, gradimy: pMatInt): Real;
var
  i: Integer;
  j: Integer;
  somi, somi1: Sommet;
  somt: TList;
  Es, E: Real;
begin
  E:= 0;
  for i:= 0 to polys.Sommets.Count-2 do begin
    somi:= Sommet(polys.Sommets[i]);
    somi1:= Sommet(polys.Sommets[i+1]);
    if (somi.BordU and somi1.BordU and (somi.U = somi1.U)) or (somi.BordV and somi1.BordV and (somi.V = somi1.V)) then Continue
    else begin
      { l'energie de ce segment a-t-elle deja ete calculee ? }

      Es:= ChercheE_Segments(somi.U, somi.V, somi1.U, somi1.V);
      if (Es <> -100) then begin
        E:= E + Es;
      end
      else begin
        { on calcule et on stocke }
        somt:= echantSegment(somi, somi1, pas);
        Es:= EnergieSegment(somt, gradimx, gradimy);
        E:= E + Es;
        StockeE_Segments(somi.U, somi.V, somi1.U, somi1.V, Es);
        { liberation des points d'echantillonage
          sauf le premier qui est un point du polygone courant }
        for j:= 1 to somt.Count-1 do
          Sommet(somt[j]).Free; 
      end;
    end;
  end;
  Result:= E;
end;

function InterpBicubique(u, v: Real; pmat: pMatInt): Real;
var
  j,k:integer;
  //a,b:real;
  s, t, t0, t1, t2:real;
  RC: TMatInt;
begin
  { t -> lignes, s -> colonnes }
  RC:= pmat^;
  j:= Floor(u);
  k:= Floor(v);
  t:= u-j;
  s:= v-k;

  t0:= ((2-t)*t-1)*t;
  t1:= (3*t-5)*t*t+2;
  t2:= ((4-3*t)*t+1)*t;
  t:= (t-1)*t*t;

  Result:= (
           (RC[j-1,k-1]*t0 + RC[j,k-1]*t1 + RC[j+1,k-1]*t2 + RC[j+2,k-1]*t)*(((2-s)*s-1)*s)+
           (RC[j-1,k]*t0 + RC[j,k]*t1 + RC[j+1,k]*t2 + RC[j+2,k]*t)*((3*s-5)*s*s+2)+
           (RC[j-1,k+1]*t0 + RC[j,k+1]*t1 + RC[j+1,k+1]*t2 + RC[j+2,k+1]*t)*(((4-3*s)*s+1)*s)+
           (RC[j-1,k+2]*t0 + RC[j,k+2]*t1 + RC[j+1,k+2]*t2 + RC[j+2,k+2]*t)*((s-1)*s*s)
           ) / 4;
end;

function EnergieSegment(points: TList; pgradx, pgrady: pMatInt): Real;
var
  pente, e: Sommet;
  r, np, Vx, Vy, V, En: Real;
  i, N: Integer;
begin
  { calcul de la pente du segment }
  pente:= Sommet.Create(Sommet(points[points.Count-1]).V-Sommet(points[0]).V, Sommet(points[points.Count-1]).U-Sommet(points[0]).U);
  r:= -pente.V;
  pente.V:= pente.U;
  pente.U:= r;
  np:= NormVect(Sommet.Create, pente);
  pente.U:= abs(pente.U / np);
  pente.V:= abs(pente.V / np);

  { on enlve le dernier point du segment, il sera pris en compte comme
    premier point du segment suivant }
   points.Delete(points.Count-1);

  { on somme les gradients interpols des points }
  N:= 0;
  En:= 0;
  for i:= 0 to points.Count-1 do begin
    { le point ne doit pas tre sur un bord de l'image car pas de
      gradient pour lui }
    e:= Sommet(points[i]);
    { on retire les points du segment qui sont sur un bord de l'image
      ces points n'ont pas de gradient }
    if ((e.U < 1) or (e.U >= (NLig-1)) or (e.V < 1) or (e.V >= (NCol-1))) then Continue;
    { interpolation }
    Vx:= InterpBicubique(e.U, e.V, pgradx);
    Vy:= InterpBicubique(e.U, e.V, pgrady);
    V:= Vx * pente.U + Vy * pente.V;
//FMain.Memo1.Lines.Add( FloatToStr(V*V));
//showmessage('pause');
    En:= En + V*V;
    Inc(N);
  end;
  if (N <> 0) then
    Result:= En/N
  else
    Result:= 0;

  pente.Free;
end;

function NormVect(Sommet1, Sommet2: Sommet): Real;
begin
  Result:= Sqrt(Sqr(Sommet2.U-Sommet1.U) + Sqr(Sommet2.V-Sommet1.V));
end;

function TriUC(Item1, Item2: Pointer): Integer;
var
  U1, U2: Real;
begin
  U1:= Sommet(Item1).U;
  U2:= Sommet(Item2).U;
  if (U1 < U2) then Result:= -1
  else if (U1 > U2) then Result:= 1
  else Result:= 0;
end;

function TriVC(Item1, Item2: Pointer): Integer;
var
  V1, V2: Real;
begin
  V1:= Sommet(Item1).V;
  V2:= Sommet(Item2).V;
  if (V1 < V2) then Result:= -1
  else if (V1 > V2) then Result:= 1
  else Result:= 0;
end;

function TriUD(Item1, Item2: Pointer): Integer;
var
  U1, U2: Real;
begin
  U1:= Sommet(Item1).U;
  U2:= Sommet(Item2).U;
  if (U1 > U2) then Result:= -1
  else if (U1 < U2) then Result:= 1
  else Result:= 0;
end;

function TriVD(Item1, Item2: Pointer): Integer;
var
  V1, V2: Real;
begin
  V1:= Sommet(Item1).V;
  V2:= Sommet(Item2).V;
  if (V1 > V2) then Result:= -1
  else if (V1 < V2) then Result:= 1
  else Result:= 0;
end;

procedure ConcatListe(Liste1, Liste2: TList);
var
  i: Integer;
begin
  for i:= 0 to Liste2.Count-1 do
    Liste1.Add(Liste2.Items[i]);
end;

function EchantSegment(Sommet1, Sommet2: Sommet; pas: Real): TList;
var
  sv1, sv2: Real;
  se, st: TList;
  i: Integer;
  s: Sommet;
begin
  { init }
  st:= TList.Create;
  se:= TList.Create;
  se.Add(Sommet1);
  se.Add(Sommet2);

  { sens de variation du segment }
  sv1:= Sommet2.U - Sommet1.U;
  sv2:= Sommet2.V - Sommet1.V;

  while (NormVect(Sommet(se[0]), Sommet(se[1])) > pas) do begin
    { subdivision }
    for i:= 0 to se.Count-2 do begin
      s:= Sommet.Create;
      s.U:= (Sommet(se[i]).U + Sommet(se[i+1]).U) / 2;
      s.V:= (Sommet(se[i]).V + Sommet(se[i+1]).V) / 2;
      st.Add(s);
    end;

    { ajout des nouveaux points st  se }
    ConcatListe(se, st);
    st.Clear;

    { tri des sommets de se selon les U croissants ou non }
    if (sv1 < 0) then
      se.Sort(TriUD)
    else if (sv1 > 0) then
      se.Sort(TriUC)
    else begin
      if (sv2 < 0) then
        se.Sort(TriVD)
      else
        se.Sort(TriVC);
    end;
  end;
  Result:= se;
end;

function Gradient(image: TBitmap; const mask: array of integer): pMatInt;
var
  mat: TMatInt;
  w, h, i, j: Integer;
  rowu, rowc, rowd: pByteArray;
begin
  { taille plus grande car extension }
  w:= image.Width + 1;
  h:= image.Height + 1;
  { init memoire de la matrice }
  Result:= New(pMatInt);
  setLength(mat, h);
  for i:= 0 to h-1 do
    setLength(mat[i], w);

  { convolution }
  for i:= 1 to h-3 do begin
    rowu:= image.ScanLine[i-1];
    rowc:= image.ScanLine[i];
    rowd:= image.ScanLine[i+1];
    for j:= 1 to w-3 do begin
      //showmessage(IntToStr(rowc[j]));
      mat[i][j]:= abs(((rowu[j-1]*mask[8]) +
      (rowu[j]*mask[7]) + (rowu[j+1]*mask[6]) +
      (rowc[j-1]*mask[5]) +
      (rowc[j]*mask[4]) + (rowc[j+1]*mask[3])+
      (rowd[j-1]*mask[2]) +
      (rowd[j]*mask[1]) + (rowd[j+1]*mask[0])));
    end;
  end;

  { remplissage des lignes restes vides }
  for j:= 0 to w-1 do begin
    mat[0][j]:= mat[1][j];
    mat[h-2][j]:= mat[h-3][j];
    mat[h-1][j]:= mat[h-3][j];
  end;
  for i:= 0 to h-1 do begin
    mat[i][0]:= mat[i][1];
    mat[i][h-2]:= mat[i][h-3];
    mat[i][h-1]:= mat[i][h-3];
  end;

  Result^:= mat;
end;

constructor Sommet.Create(U, V: Real);
begin
  Self.U:= U;
  Self.V:= V;
  Self.BordU:= false;
  Self.BordV:= false;
  Self.Bloque:= false;
  Self.index:= -1;
end;

constructor Sommet.Create;
begin
  Self.U:= 0;
  Self.V:= 0;
  Self.BordU:= false;
  Self.BordV:= false;
  Self.Bloque:= false;
  Self.Index:= -1;
end;

constructor Polygone.Create;
begin
  Sommets:= TList.Create;
  Locked:= false;
  inherited Create;
end;

destructor Polygone.Destroy;
begin
  Sommets.Free;
  inherited Destroy;
end;

procedure Polygone.FreeSommets;
var
  i: Integer;
begin
  for i:= 0 to Sommets.Count-1 do
    Sommet(Sommets[i]).Free;
end;

procedure Polygone.Clone(p: Polygone);
var
  i: Integer;
begin
  Sommets.Clear;
  for i:= 0 to p.Sommets.Count-1 do begin
    Sommets.Add(Sommet(p.Sommets[i]).Clone);
  end;
end;

procedure Polygone.CopyUVFrom(p: Polygone);
var
  i: Integer;
begin
  for i:= 0 to p.Sommets.Count-1 do begin
    Sommet(Sommets[i]).U:= Sommet(p.Sommets[i]).U;
    Sommet(Sommets[i]).V:= Sommet(p.Sommets[i]).V;
  end;
end;

function Sommet.Clone: Sommet;
begin
  Result:= Sommet.Create(Self.U, Self.V);
  Result.BordU:= Self.BordU;
  Result.BordV:= Self.BordV;
  Result.Bloque:= Self.Bloque;
end;

end.
