unit image;

interface

uses
  SysUtils, Types, Classes, QGraphics, QControls, QForms, QDialogs,
  QStdCtrls, QExtCtrls, QImgList, fonctions;

type
  TFImage = class(TForm)
    PaintBox: TPaintBox;
    ImageList1: TImageList;
    procedure PaintBoxPaint(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure PaintBoxMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure PaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormActivate(Sender: TObject);
    procedure PaintBoxMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    { Private declarations }
    TheBuffer: TBitmap;
//    BackBitmap: TBitmap;
    ImagesLoaded: Boolean;
    Title: string;
    procedure PaintZone(index: Integer);
  public
    { Public declarations }
    TheImage: TBitmap;
    ZoomFact: Integer;
    Polygones: TList;
    Etat: Integer;
    SomRef, TargetRef, MoveRef: Sommet;
    function Loadmage(location: string; dir: string; name: string): Boolean;
    procedure FitImage;
    procedure PaintBuffer;
    procedure Zoom(N, X, Y: integer);
    procedure ZoomPlus(X, Y: Integer);
    procedure ZoomMoins(X, Y: Integer);
    procedure RefreshBox;
    function TX(X: Real): Integer;
    function TY(Y: Real): Integer;
    function ITX(X: Real): Integer;
    function ITY(Y: Real): Integer;
  end;

var
  FImage: TFImage;

const
  ZOOMMAX = 10;
  CROIXW = 6;
  AIM = 4;
  ZONEW = 24;
  PTSW = 3;

implementation

uses main, math;

{$R *.xfm}

procedure TFImage.PaintZone(index: Integer);
var
  U, V: Integer;
  sup: Boolean;
begin
  sup:= (Ys > (PaintBox.Height/2));
  U:= TY(ITY(Ys));
  if (sup) then
    U:= U - ZONEW;

  sup:= (Xs > (PaintBox.Width/2));
  V:= TX(ITX(Xs));
  if (sup) then
    V:= V - ZONEW;
    
  ImageList1.Draw(TheBuffer.Canvas, V, U, index);
end;

function TFImage.TX(X: Real): Integer;
begin
  Result:= Round(X*ZoomFact - HorzScrollBar.Position);
end;

function TFImage.TY(Y: Real): Integer;
begin
  Result:= Round(Y*ZoomFact - VertScrollBar.Position);
end;

function TFImage.ITX(X: Real): Integer;
begin
  Result:= Floor(X/ZoomFact);
end;

function TFImage.ITY(Y: Real): Integer;
begin
  Result:= Floor(Y/ZoomFact);
end;

function TFimage.Loadmage(location: string; dir: string; name: string): Boolean;
begin
  Result:= true;
  try
    TheImage.LoadFromFile(location);
    FitImage;
    Title:= 'Image ' + dir + ': ' + name;
    Caption:= Title + ' [1:1]';
    ImagesLoaded:= true;
    PaintBuffer;
    RefreshBox;
  except
    Result:= false;
    Exit;
  end;
end;

procedure TFImage.FitImage;
begin
  Width:= TheImage.Width;
  Height:= TheImage.Height;
  PaintBox.Width:= TheImage.Width;
  PaintBox.Height:= TheImage.Height;
end;

procedure TFImage.PaintBoxPaint(Sender: TObject);
begin
  if (ImagesLoaded) then begin
    PaintBuffer;
    RefreshBox;
  end;
end;

procedure TFImage.FormCreate(Sender: TObject);
begin
  Etat:= 0;
  TheBuffer:= TBitmap.Create;
  TheBuffer.Width:= 1024;
  TheBuffer.Height:= 1024;
  TheImage:= TBitmap.Create;
  ZoomFact:= 1;
  ImagesLoaded:= false;
  Polygones:= TList.Create;
  SomRef:= nil;
  TargetRef:= nil;
  MoveRef:= nil;
end;

procedure TFImage.PaintBuffer;
var
  tmpbitmap: TBitmap;
  p, i, Xi, Xf, Yi, Yf, X, Y: Integer;
  poly: Polygone;
  som: Sommet;
  points: array of TPoint;
begin
  Xi:= Floor(HorzScrollbar.Position/ZoomFact);
  Yi:= Floor(VertScrollbar.Position/ZoomFact);

  tmpbitmap:= TBitmap.Create;
  tmpbitmap.Width:= Floor(Width/ZoomFact)+1;
  tmpbitmap.Height:= Floor(Height/ZoomFact)+1;
  tmpbitmap.Canvas.CopyRect(Bounds(0, 0, tmpbitmap.Width, tmpbitmap.Height), TheImage.Canvas, Bounds(Xi, Yi, tmpbitmap.Width, tmpbitmap.Height));

  Xf:= Xi+tmpbitmap.Width;
  Yf:= Yi+tmpbitmap.Height;

  //TheBuffer.Canvas.StretchDraw(Bounds(0, 0, ZoomFact*TheImage.Width, ZoomFact*TheImage.Height), TheImage);
  TheBuffer.Canvas.StretchDraw(Bounds(0, 0, ZoomFact*tmpbitmap.Width, ZoomFact*tmpbitmap.Height), tmpbitmap);
  //TheBuffer.Canvas.CopyRect(bounds(DemiCarre,DemiCarre,Wzf,Hzf), infos.backbitmap.Canvas, bounds(0,0,Wzf,Hzf));
  tmpbitmap.Free;

  { les polygones et les sommets }
  TheBuffer.Canvas.Pen.Color:= clyellow;
  TheBuffer.Canvas.Brush.Color:= clyellow;
  TheBuffer.Canvas.Brush.Style:= bsSolid;
  for p:= 0 to Polygones.Count-1 do begin
    poly:= Polygone(Polygones[p]);
    { trac du polygone }
    SetLength(points, poly.Sommets.Count);
    for i:= 0 to poly.Sommets.Count-1 do begin
      som:= Sommet(poly.Sommets[i]);
      points[i]:= Point(TX(som.V), TY(som.U));
      ///
      if (points[i].X = TheImage.Width) then
        points[i].X:= TheImage.Width-1;
      if (points[i].Y = TheImage.Height) then
        points[i].Y:= TheImage.Height-1;
    end;
    TheBuffer.Canvas.Polyline(points, 0, -1);

    { trac des sommets }
    TheBuffer.Canvas.Pen.Color:= clyellow;
    for i:= 0 to poly.Sommets.Count-1 do begin
      som:= Sommet(poly.Sommets[i]);
      X:= Round(som.V);
      Y:= Round(som.U);
      if ((X > Xf) or (X < Xi) or (Y > Yf) or (Y < Yi)) then Continue;
        X:= TX(som.V);
        Y:= TY(som.U);
        { la croix pour un sommet }
        TheBuffer.Canvas.Polyline([Point(X-CROIXW,Y), Point(X-1,Y)], 0, -1);
        TheBuffer.Canvas.Polyline([Point(X+1,Y), Point(X+CROIXW,Y)], 0, -1);
        TheBuffer.Canvas.Polyline([Point(X,Y-CROIXW), Point(X,Y-1)], 0, -1);
        TheBuffer.Canvas.Polyline([Point(X,Y+1), Point(X,Y+CROIXW)], 0, -1);
    end;
  end;

  { la ligne en cours }
  if (Mode = 1) then begin
    if (Etat = 1) then begin
      TheBuffer.Canvas.Polyline([Point(TX(SomRef.V),TY(SomRef.U)), Point(TX(ITX(Xs)),TY(ITY(Ys)))], 0, -1);
    end;
  end;

  { les points d'intrts }
  if (PtsIntDrawable and (Self = ImageG)) then begin
    TheBuffer.Canvas.Pen.Color:= clred;
    TheBuffer.Canvas.Brush.Style:= bsSolid;
    for p:= 0 to ImageG.Polygones.Count-1 do begin
      if (PtsInt[p] <> nil) then begin
        for i:= 0 to Length(PtsInt[p])-1 do begin
          if (PtsInt[p][i].val = -100) then Continue;
          X:= Round(PtsInt[p][i].V);
          Y:= Round(PtsInt[p][i].U);
          if ((X > Xf) or (X < Xi) or (Y > Yf) or (Y < Yi)) then Continue;
          X:= TX(X);
          Y:= TY(Y);
          TheBuffer.Canvas.PenPos:= Point(X-PTSW, Y);
          TheBuffer.Canvas.LineTo(X+PTSW, Y);
          TheBuffer.Canvas.PenPos:= Point(X, Y-PTSW);
          TheBuffer.Canvas.LineTo(X, Y+PTSW);
        end;
      end;
    end;
  end;

  { les points d'intrts apparis }
  if (PtsIntAppDrawable) then begin
    TheBuffer.Canvas.Pen.Color:= clblue;
    TheBuffer.Canvas.Brush.Style:= bsSolid;
    for p:= 0 to Length(PtsIntApp)-1 do begin
      for i:= 0 to Length(PtsIntApp[p])-1 do begin
        if (PtsIntApp[p][i].val = -100) then Continue;
        if (Self = ImageG) then begin
          X:= Round(PtsIntApp[p][i].j);
          Y:= Round(PtsIntApp[p][i].i);
        end
        else begin
          X:= Round(PtsIntApp[p][i].V);
          Y:= Round(PtsIntApp[p][i].U);
        end;
        if ((X > Xf) or (X < Xi) or (Y > Yf) or (Y < Yi)) then Continue;
        X:= TX(X);
        Y:= TY(Y);
        TheBuffer.Canvas.PenPos:= Point(X-PTSW, Y);
        TheBuffer.Canvas.LineTo(X+PTSW, Y);
        TheBuffer.Canvas.PenPos:= Point(X, Y-PTSW);
        TheBuffer.Canvas.LineTo(X, Y+PTSW);
      end;
    end;
  end;

  { les points d'intrts apparis aprs ransac }
  if (PtsIntAppRobDrawable) then begin
    TheBuffer.Canvas.Pen.Color:= clgreen;
    TheBuffer.Canvas.Brush.Style:= bsSolid;
    for p:= 0 to Length(PtsIntAppRob)-1 do begin
      for i:= 0 to Length(PtsIntAppRob[p])-1 do begin
        if (PtsIntAppRob[p][i].val = -100) then Continue;
        if (Self = ImageG) then begin
          X:= Round(PtsIntAppRob[p][i].j);
          Y:= Round(PtsIntAppRob[p][i].i);
        end
        else begin
          X:= Round(PtsIntAppRob[p][i].V);
          Y:= Round(PtsIntAppRob[p][i].U);
        end;
        if ((X > Xf) or (X < Xi) or (Y > Yf) or (Y < Yi)) then Continue;
        X:= TX(X);
        Y:= TY(Y);
        TheBuffer.Canvas.PenPos:= Point(X-PTSW, Y);
        TheBuffer.Canvas.LineTo(X+PTSW, Y);
        TheBuffer.Canvas.PenPos:= Point(X, Y-PTSW);
        TheBuffer.Canvas.LineTo(X, Y+PTSW);
      end;
    end;
  end;

  { l'indicateur de zone particulire }
  //FMain.Edit1.Text:= FloatToStr(Uz)+' '+FloatToStr(Vz);
  if (TargetRef <> nil) then begin
    PaintZone(0);
  end
  else if ((Uz <> -1) and (Vz <> -1)) then begin
    PaintZone(3);
  end
  else if (Uz <> -1) then begin
    PaintZone(1);
  end
  else if (Vz <> -1) then begin
    PaintZone(2);
  end;
end;

procedure TFImage.RefreshBox;
begin
  PaintBox.Canvas.Draw(HorzScrollbar.Position, VertScrollbar.Position, TheBuffer);
  //PaintBox.Canvas.CopyRect(Bounds(HorzScrollbar.Position, VertScrollbar.Position, Width, Height), TheBuffer.Canvas, Bounds(0, 0, Width, Height));
end;

procedure TFImage.FormDestroy(Sender: TObject);
var
  i: Integer;
begin
  TheBuffer.Free;
  TheImage.Free;
  for i:= 0 to Polygones.Count-1 do begin
//    Polygone(Polygones[i]).FreeSommets;
    Polygone(Polygones[i]).Free;
  end;
  Polygones.Free;
end;

procedure TFImage.PaintBoxMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  i, j: Integer;
  poly: Polygone;
  som: Sommet;
  found: Boolean;
  Xi, Yi: Integer;
begin
  FMain.LabelUV.Caption:= '(u,v) = ('+IntToStr(Floor(Y/ZoomFact))+','+IntToStr(Floor(X/ZoomFact))+')';
  Xs:= X;
  Ys:= Y;
  Uz:= -1;
  Vz:= -1;
  if (Mode = 1) then begin
    { tester si on est dans une zone particulire }
    Xi:= ITX(X);
    Yi:= ITY(Y);
    //Uz:= -1;
    //Vz:= -1;
    { une zone sommet est prioritaire sur les bords }
    found:= false;
    TargetRef:= nil;
    for i:= 0 to Polygones.Count-1 do begin
      poly:= Polygone(Polygones[i]);
      for j:= 0 to poly.Sommets.Count-1 do begin
        som:= Sommet(poly.Sommets[j]);
        if ((som.V > (Xi-AIM)) and (som.V < (Xi+AIM)) and (som.U > (Yi-AIM)) and (som.U < (Yi+AIM))) then begin
          Uz:= som.U;
          Vz:= som.V;
          found:= true;
          TargetRef:= som;
          Break;
        end;
      end;
      if (found) then Break;
    end;
    if (not found) then begin
      if (Xi < AIM) then
        Vz:= 0
      else if (Xi > (TheImage.Width-1-AIM)) then
        Vz:= TheImage.Width-1;
      if (Yi < AIM) then
        Uz:= 0
      else if (Yi > (TheImage.Height-1-AIM)) then
        Uz:= TheImage.Height-1;
    end;
  end
  else if (Mode = 2) then begin
    { tester si on est prs d'un sommet }
    Xi:= ITX(X);
    Yi:= ITY(Y);
    if (Etat = 0) then begin
      found:= false;
      TargetRef:= nil;
      for i:= 0 to Polygones.Count-1 do begin
        poly:= Polygone(Polygones[i]);
        for j:= 0 to poly.Sommets.Count-1 do begin
          som:= Sommet(poly.Sommets[j]);
          if ((som.V > (Xi-AIM)) and (som.V < (Xi+AIM)) and (som.U > (Yi-AIM)) and (som.U < (Yi+AIM))) then begin
            Uz:= som.U;
            Vz:= som.V;
            found:= true;
            TargetRef:= som;
            Break;
          end;
        end;
        if (found) then Break;
      end;
    end
    else if (Etat = 1) then begin
      if (MoveRef <> nil) then begin
        MoveRef.U:= Yi;
        MoveRef.V:= Xi;
      end;
      { tester les bords }
      if (Xi < AIM) then
        Vz:= 0
      else if (Xi > (TheImage.Width-1-AIM)) then
        Vz:= TheImage.Width-1;
      if (Yi < AIM) then
        Uz:= 0
      else if (Yi > (TheImage.Height-1-AIM)) then
        Uz:= TheImage.Height-1;
      FMain.PolygonesToMemo;
    end
  end
  else if (Mode = 3) then begin
    if (Etat = 0) then begin
      Xi:= ITX(X);
      Yi:= ITY(Y);
      { on cherche un sommet }
      found:= false;
      TargetRef:= nil;
      for i:= 0 to Polygones.Count-1 do begin
        poly:= Polygone(Polygones[i]);
        for j:= 0 to poly.Sommets.Count-1 do begin
          som:= Sommet(poly.Sommets[j]);
          if ((som.V > (Xi-AIM)) and (som.V < (Xi+AIM)) and (som.U > (Yi-AIM)) and (som.U < (Yi+AIM))) then begin
            Uz:= som.U;
            Vz:= som.V;
            found:= true;
            TargetRef:= som;
            Break;
          end;
        end;
        if (found) then Break;
      end;
    end;
  end
  else if (Mode = 4) then begin
    { tester si on est prs d'un sommet }
    Xi:= ITX(X);
    Yi:= ITY(Y);
    found:= false;
    TargetRef:= nil;
    for i:= 0 to Polygones.Count-1 do begin
      poly:= Polygone(Polygones[i]);
      for j:= 0 to poly.Sommets.Count-1 do begin
        som:= Sommet(poly.Sommets[j]);
        if ((som.V > (Xi-AIM)) and (som.V < (Xi+AIM)) and (som.U > (Yi-AIM)) and (som.U < (Yi+AIM))) then begin
          Uz:= som.U;
          Vz:= som.V;
          found:= true;
          TargetRef:= som;
          Break;
        end;
      end;
      if (found) then Break;
    end;
  end;
  PaintBuffer;
  RefreshBox;
end;

procedure TFImage.Zoom(N, X, Y: integer);
var
  X0, Y0: Integer;
begin
  X0:= Round((N/ZoomFact-1)*X) + HorzScrollBar.Position;
  Y0:= Round((N/ZoomFact-1)*Y) + VertScrollBar.Position;
  ZoomFact:= N;
  Caption:= Title + ' [' + IntToStr(ZoomFact) + ':1]';
  if (X0 < 0) then X0:= 0;
  if (Y0 < 0) then Y0:= 0;

  if (ZoomFact = 1) then begin
    FitImage;
  end
  else begin
    PaintBox.Width:= TheImage.Width*N;
    PaintBox.Height:= TheImage.Height*N;
    HorzScrollBar.Position:= X0;
    VertScrollBar.Position:= Y0;
  end;
  PaintBuffer;
  RefreshBox;
end;

procedure TFImage.Button1Click(Sender: TObject);
begin
  ZoomPlus(Round(Width/2), Round(Height/2));
//  FMain.Edit1.Text:= IntToStr(Width)+' '+IntToStr(Height)
end;

procedure TFImage.ZoomPlus(X, Y: Integer);
begin
  if (ZoomFact < ZOOMMAX) then begin
    Zoom(ZoomFact+1, X, Y);
  end;
end;

procedure TFImage.ZoomMoins(X, Y: Integer);
begin
  if (ZoomFact > 1) then begin
    Zoom(ZoomFact-1, X, Y);
  end;
end;

procedure TFImage.Button2Click(Sender: TObject);
begin
  ZoomMoins(Round(Width/2), Round(Height/2));
end;

procedure TFImage.PaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  poly: Polygone;
  som: Sommet;
  i, n: Integer;
  s1, s2, s3, s4, c1, c2, Inter: array[0..2] of Real;
begin
  if (ssShift in Shift) then begin
    if (Button = mbLeft) then ZoomPlus(X, Y)
    else if (Button = mbMiddle) then ZoomPlus(X, Y)
    else if (Button = mbRight) then ZoomMoins(X, Y);
  end
  else begin
    if (Mode = 1) then begin
      if (Etat = 0) then begin
        if (Button = mbLeft) then begin
          poly:= Polygone.Create;
          { test si sommet particulier }
          if (TargetRef <> nil) then begin
            poly.Sommets.Add(TargetRef);
            SomRef:= TargetRef;
          end
          else if ((Uz <> -1) and (Vz <> -1)) then begin
            som:= Sommet.Create(Uz,Vz);
            som.BordU:= true;
            som.BordV:= true;
            poly.Sommets.Add(som);
            SomRef:= som;
          end
          else if (Uz <> -1) then begin
            som:= Sommet.Create(Uz,ITY(X));
            som.BordU:= true;
            poly.Sommets.Add(som);
            SomRef:= som;
          end
          else if (Vz <> -1) then begin
            som:= Sommet.Create(ITX(Y),Vz);
            som.BordV:= true;
            poly.Sommets.Add(som);
            SomRef:= som;
          end
          else begin
            som:= Sommet.Create(ITX(Y),ITY(X));
            poly.Sommets.Add(som);
            SomRef:= som;
          end;
          Etat:= 1;
          Polygones.Add(poly);
          FMain.PolygonesToMemo;
        end
        else begin
        end;
      end
      else if (Etat = 1) then begin
        if (Button = mbLeft) then begin
          if (TargetRef <> nil) then begin
            Polygone(Polygones[Polygones.Count-1]).Sommets.Add(TargetRef);
            SomRef:= TargetRef;
            if (Polygone(Polygones[Polygones.Count-1]).Sommets[0] = Polygone(Polygones[Polygones.Count-1]).Sommets[Polygone(Polygones[Polygones.Count-1]).Sommets.Count-1]) then begin
              Etat:= 0;
              SomRef:= nil;
            end;
          end
          else if ((Uz <> -1) and (Vz <> -1)) then begin
            som:= Sommet.Create(Uz,Vz);
            som.BordU:= true;
            som.BordV:= true;
            Polygone(Polygones[Polygones.Count-1]).Sommets.Add(som);
            SomRef:= som;
          end
          else if (Uz <> -1) then begin
            som:= Sommet.Create(Uz,ITY(X));
            som.BordU:= true;
            Polygone(Polygones[Polygones.Count-1]).Sommets.Add(som);
            SomRef:= som;
          end
          else if (Vz <> -1) then begin
            som:= Sommet.Create(ITX(Y),Vz);
            som.BordV:= true;
            Polygone(Polygones[Polygones.Count-1]).Sommets.Add(som);
            SomRef:= som;
          end
          else begin
            som:= Sommet.Create(ITX(Y),ITY(X));
            Polygone(Polygones[Polygones.Count-1]).Sommets.Add(som);
            SomRef:= som;
          end;
          FMain.PolygonesToMemo;
        end
        else begin
          Etat:= 0;
        end;
      end;
    end
    else if (Mode = 2) then begin
      if (Etat = 0) then begin
        if (Button = mbLeft) then begin
          if (TargetRef <> nil) then begin
            MoveRef:= TargetRef;
            TargetRef:= nil;
            Cursor:= crNone;
            Etat:= 1;
          end;
        end;
      end;
    end
    else if (Mode = 3) then begin
      if (Etat = 0) then begin
        if (TargetRef <> nil) then begin
          for i:= 0 to Polygones.Count-1 do begin
            poly:= Polygone(Polygones[i]);
            n:= 0;
            while(n < poly.Sommets.Count) do begin
              if (Sommet(poly.Sommets[n]) = TargetRef) then begin
                poly.Sommets.Delete(n);
              end
              else
                Inc(n);
            end;
          end;
          TargetRef.Free;
          TargetRef:= nil;
          { enlever les polygones vides }
          n:= 0;
          while (n < Polygones.Count) do begin
            if (Polygone(Polygones[n]).Sommets.Count = 0) then
              Polygones.Delete(n)
            else
              Inc(n);
          end;
        end;
      end;
    end
    else if (Mode = 4) then begin
      if (TargetRef <> nil) then begin
        Sommets_ajust[Etat]:= TargetRef;
        Etat:= Etat + 1;
        if (Etat = 4) then begin
          s1[0]:= Sommets_ajust[0].V;
          s1[1]:= Sommets_ajust[0].U;
          s1[2]:= 1;
          s2[0]:= Sommets_ajust[1].V;
          s2[1]:= Sommets_ajust[1].U;
          s2[2]:= 1;
          s3[0]:= Sommets_ajust[2].V;
          s3[1]:= Sommets_ajust[2].U;
          s3[2]:= 1;
          s4[0]:= Sommets_ajust[3].V;
          s4[1]:= Sommets_ajust[3].U;
          s4[2]:= 1;
          { les deux cotes }
          ProdVect(s1,s2,c1);
          NormaliseHomogene2D(c1);
          ProdVect(s3,s4,c2);
          NormaliseHomogene2D(c2);
          ProdVect(c1,c2,Inter);
          NormaliseHomogene2D(Inter);
          Etat:= 0;
          Sommets_ajust[1].U:= Inter[1];
          Sommets_ajust[1].V:= Inter[0];
          
        end;
        TargetRef:= nil;
      end;
    end;
  end;
  PaintBuffer;
  RefreshBox;
  FMain.PolygonesToMemo;
end;

procedure TFImage.FormActivate(Sender: TObject);
begin
  FMain.ActiveImage:= Self;
end;

procedure TFImage.PaintBoxMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if (Mode = 2) then begin
    if (Etat = 1) then begin
      if (Uz <> -1) then MoveRef.U:= Uz;
      if (Vz <> -1) then MoveRef.V:= Vz;
      MoveRef:= nil;
      Etat:= 0;
      Cursor:= crDefault;
    end;
  end;
end;

end.


