unit polyfunc;

interface

uses
  SysUtils, QGraphics, QDialogs, Classes, types, matrices, fonctions;


function G_PtInPolygon(PolyPoints: array of TRoint; ptTest: TRoint):
Boolean;

implementation


{*************************************************************************
 * FUNCTION:   CCW (CounterClockWise)
 *
 * PURPOSE
 * Determines, given three points, if when travelling from the first to
 * the second to the third, we travel in a counterclockwise direction.

 *
 * RETURN VALUE
 * (int) 1 if the movement is in a counterclockwise direction, -1 if
 * not.
 *************************************************************************}

function CCW(p0,p1,p2:TRoint): Integer;
var
  dx1, dx2: Real;
  dy1, dy2: Real;
begin
  dx1 := p1.x - p0.x ;
  dx2 := p2.x - p0.x ;
  dy1 := p1.y - p0.y ;
  dy2 := p2.y - p0.y ;

  {* This is basically a slope comparison: we don't do divisions because
   * of divide by zero possibilities with pure horizontal and pure
   * vertical lines.
   *}
  if((dx1 * dy2) > (dy1 * dx2))then
    Result := 1
  else
    Result := -1;
end;

{*************************************************************************
 * FUNCTION:   Intersect
 *
 * PURPOSE
 * Given two line segments, determine if they intersect.
 *
 * RETURN VALUE
 * TRUE if they intersect, FALSE if not.
 *************************************************************************}

function Intersect(p1,p2,p3,p4: TRoint): Boolean;
begin
  Result := (((CCW(p1, p2, p3) * CCW(p1, p2, p4)) <= 0)and
            (( CCW(p3, p4, p1) * CCW(p3, p4, p2)  <= 0))) ;
end;

{*************************************************************************
 * FUNCTION:   G_PtInPolyRect
 *
 * PURPOSE
 * This routine determines if a point is within the smallest rectangle
 * that encloses a polygon.
 *
 * RETURN VALUE
 * (BOOL) TRUE or FALSE depending on whether the point is in the rect or

 * not.
 *************************************************************************}

function G_PtInPolyRect(PolyPoints: array of TRoint; ptTest: TRoint; var
prbound: TRRect): Boolean;
var
  xmin, xmax, ymin, ymax: Real;
  pt: TRoint;
  i: Word;
begin
  xmin := MaxInt;
  ymin := MaxInt;
  xmax := -MaxInt;
  ymax := -MaxInt;

  for i := 0 to High(PolyPoints) do
  begin
    pt := PolyPoints[i];
    if (pt.x < xmin)then
      xmin := pt.x;
    if (pt.x > xmax)then
      xmax := pt.x;
    if (pt.y < ymin)then
      ymin := pt.y;
    if (pt.y > ymax)then
      ymax := pt.y;
   end;
   prbound := RRect(xmin, ymin, xmax, ymax);
   Result := PtInRRect(prbound,ptTest);
end;

{*************************************************************************
 * FUNCTION:   G_PtInPolygon
 *
 * PURPOSE
 * This routine determines if the point passed is in the polygon. It uses

 * the classical polygon hit-testing algorithm: a horizontal ray starting

 * at the point is extended infinitely rightwards and the number of
 * polygon edges that intersect the ray are counted. If the number is odd,

 * the point is inside the polygon.
 *
 * RETURN VALUE
 * (BOOL) TRUE if the point is inside the polygon, FALSE if not.
 *************************************************************************}

function G_PtInPolygon(PolyPoints: array of TRoint; ptTest: TRoint):
Boolean;
var
  i: Integer;
  pt1, pt2, pt: TRoint;
  wnumintsct: Word;
  prbound: TRRect;
begin
  wnumintsct := 0;
  Result := False;
  { si le point n'est pas dans le rectangle englobant alors il n'est pas dans
    le polygone }
  if(not G_PtInPolyRect(PolyPoints,ptTest,prbound))then
    Exit;

  { si le point est sur le polygone alors il est considr comme dedans }
  for i := 0 to High(PolyPoints) do begin
    if (i <> High(PolyPoints)) then
      pt:= PolyPoints[i+1]
    else
      pt:= PolyPoints[0];
    { a et b }
    pt1.x:= pt.y - PolyPoints[i].y;
    pt1.y:= PolyPoints[i].x - pt.x;
    { c }
    pt2.x:= -pt1.x*PolyPoints[i].x - pt1.y*PolyPoints[i].y;
    if (abs(pt1.x*ptTest.x + pt1.y*ptTest.y + pt2.x) < 1E-3) then begin
      Result:= true;
      Exit;
    end;
  end;

  pt1 := ptTest;
  pt2 := ptTest;
  pt2.x := prbound.Right+50;

  // Now go through each of the lines in the polygon and see if it
  // intersects
  for i := 0 to High(PolyPoints)-1 do
    if(Intersect(ptTest,pt2,PolyPoints[i],PolyPoints[i+1]))then
      Inc(wnumintsct);
  // And the last line
  if(Intersect(ptTest, pt2, PolyPoints[High(PolyPoints)],
PolyPoints[0]))then
     Inc(wnumintsct);
  // If wnumintsct is odd then the point is inside the polygon
  Result := Odd(wnumintsct);
end;

end.
